from ._air_scattering_coefficient import (
    AirScatteringCoefficientSpectrum as AirScatteringCoefficientSpectrum,
)
from ._core import Spectrum as Spectrum
from ._core import spectrum_factory as spectrum_factory
from ._interpolated import InterpolatedSpectrum as InterpolatedSpectrum
from ._multi_delta import MultiDeltaSpectrum as MultiDeltaSpectrum
from ._solar_irradiance import SolarIrradianceSpectrum as SolarIrradianceSpectrum
from ._uniform import UniformSpectrum as UniformSpectrum



from __future__ import annotations

import logging
import typing as t
import warnings

import attrs
import numpy as np
import pint
import pinttr
import xarray as xr

from .index import CKDSpectralIndex
from .spectral_set import SpectralSet
from ..attrs import documented, parse_docs
from ..constants import SPECTRAL_RANGE_MAX, SPECTRAL_RANGE_MIN
from ..quad import Quad, QuadType
from ..radprops import CKDAbsorptionDatabase
from ..units import to_quantity
from ..units import unit_context_config as ucc
from ..units import unit_registry as ureg
from ..util.misc import round_to_multiple

logger = logging.getLogger(__name__)


# ------------------------------------------------------------------------------
#                              CKD bin data classes
# ------------------------------------------------------------------------------


@parse_docs
@attrs.define(eq=False, frozen=True, slots=True)
class Bin:
    """
    A data class representing a spectral bin in CKD modes.

    Notes
    -----
    A bin is more than a spectral interval. It is associated with a
    quadrature rule.
    """

    wmin: pint.Quantity = documented(
        pinttr.field(
            units=ucc.deferred("wavelength"),
            on_setattr=None,  # frozen instance: on_setattr must be disabled
        ),
        doc="Bin lower spectral bound.\n\nUnit-enabled field "
        '(default: ucc["wavelength"]).',
        type="quantity",
        init_type="quantity or float",
    )

    wmax: pint.Quantity = documented(
        pinttr.field(
            units=ucc.deferred("wavelength"),
            on_setattr=None,  # frozen instance: on_setattr must be disabled
        ),
        doc="Bin upper spectral bound.\n\nUnit-enabled field "
        '(default: ucc["wavelength"]).',
        type="quantity",
        init_type="quantity or float",
    )

    @wmin.validator
    @wmax.validator
    def _wbounds_validator(self, attribute, value):
        if not self.wmin < self.wmax:
            raise ValueError(
                f"while validating {attribute.name}: wmin must be lower than wmax"
            )

    quad: Quad = documented(
        attrs.field(
            factory=lambda: Quad.gauss_legendre(2),
            repr=lambda x: x.str_summary,
            validator=attrs.validators.instance_of(Quad),
        ),
        doc="Quadrature rule attached to the CKD bin.",
        type=":class:`.Quad`",
    )

    @property
    def width(self) -> pint.Quantity:
        """quantity : Bin spectral width."""
        return self.wmax - self.wmin

    @property
    def wcenter(self) -> pint.Quantity:
        """quantity : Bin central wavelength."""
        return 0.5 * (self.wmin + self.wmax)

    @property
    def pretty_repr(self) -> str:
        """str : Pretty representation of the bin."""
        units = ureg.Unit("nm")
        wrange = (
            f"[{self.wmin.m_as(units):.1f}, {self.wmax.m_as(units):.1f}] {units:~P}"
        )
        quad = self.quad.pretty_repr()
        return f"{wrange}: {quad}"

    def spectral_indices(self) -> t.Generator[CKDSpectralIndex]:
        for value in self.quad.eval_nodes(interval=[0.0, 1.0]):
            yield CKDSpectralIndex(w=self.wcenter, g=value)


# ------------------------------------------------------------------------------
#                          CKD quadrature setup classes
# ------------------------------------------------------------------------------

# Number of g-points of the default quadrature
_DEFAULT_NG: int = 16


@parse_docs
@attrs.define
class QuadSpec:
    """
    Abstract base class for all quadrature specification patterns.

    Each subclass defines a strategy used to generate a spectral quadrature
    corresponding to a CKD dataset and must implement the strategy in the
    :meth:`make_quad`.
    """

    @staticmethod
    def default() -> QuadSpecFixed:
        """
        Return the default spectral quadrature (Gauss-Legendre, 16 *g*-points).
        """
        return QuadSpecFixed(n=_DEFAULT_NG, quad_type="gauss_legendre")

    @staticmethod
    def from_dict(
        value: dict[str, t.Any],
    ) -> QuadSpecFixed | QuadSpecMinError | QuadSpecErrorThreshold:
        """
        Create a quadrature specification subtype from a dictionary. The
        dictionary must have a ``type`` entry, whose value maps to a give
        quadrature specification subtype as follows:

        * ``fixed``: :class:`.QuadSpecFixed`
        * ``minimize_error``: :class:`.QuadSpecMinError`
        * ``error_threshold``: :class:`.QuadSpecErrorThreshold`

        Parameters
        ----------
        value : dict
            A dictionary mapping parameter names to their respective values.

        Returns
        -------
        QuadSpec
        """
        try:
            subtype: str = value.pop("type")
        except KeyError:
            raise ValueError("dictionary input must have a 'type' entry")

        if subtype == "fixed":
            cls = QuadSpecFixed
        elif subtype in {"minimize", "minimize_error"}:
            cls = QuadSpecMinError
        elif subtype in {"threshold", "error_threshold"}:
            cls = QuadSpecErrorThreshold
        else:
            raise ValueError(f"Unknown quadrature specification '{subtype}'")

        return cls.from_dict(value)

    @classmethod
    def convert(cls, value: t.Any) -> QuadSpec:
        """
        Attempt conversion to a :class:`.QuadSpec` instance. If `value` is a
        dictionary, it is passed to :meth:`.from_dict`; otherwise, it is left
        unchanged.
        """
        if isinstance(value, dict):
            return cls.from_dict(value)
        else:
            return value

    def make_quad(self, dataset: xr.Dataset) -> Quad:
        """
        Apply the quadrature generation strategy and generate a quadrature rule
        for a given dataset.

        Parameters
        ----------
        dataset : Dataset
            An xarray dataset following the CKD absorption data format, for
            which a quadrature rule is generated.

        Returns
        -------
        .Quad
        """
        raise NotImplementedError


@parse_docs
@attrs.define
class QuadSpecFixed(QuadSpec):
    """
    Fixed number of quadrature points [``fixed``]

    Use a fixed number of quadrature points for all bins. If the quadrature
    is specified this way, the quadrature type has to be explicitly specified
    using the ``type`` field.
    """

    n: int = documented(
        attrs.field(),
        doc="Number of quadrature points",
        type="int",
    )

    quad_type: QuadType = documented(
        attrs.field(default="gauss_legendre", converter=QuadType),
        doc="Quadrature type",
        type=".QuadType",
        init_type=".QuadType or str",
        default='"gauss_legendre"',
    )

    @staticmethod
    def from_dict(value: dict[str, t.Any]) -> QuadSpecFixed:
        # Inherit docstring
        return QuadSpecFixed(**value)

    def make_quad(self, dataset: xr.Dataset) -> Quad:
        # Inherit docstring
        return Quad.new(type=self.quad_type, n=self.n)


@parse_docs
@attrs.define
class QuadSpecMinError(QuadSpec):
    """
    Error-minimizing number of quadrature points [``minimize_error``]

    Find the number of quadrature points that minimizes the error on the
    atmospheric transmittance.
    """

    nmax: int | None = documented(
        attrs.field(
            default=None,
            converter=attrs.converters.optional(int),
        ),
        doc="Maximum number of quadrature points",
        type="int or None",
        init_type="int, optional",
    )

    @staticmethod
    def from_dict(value: dict[str, t.Any]) -> QuadSpecMinError:
        return QuadSpecMinError(**value)

    def make_quad(self, dataset: xr.Dataset) -> Quad:
        # Inherit docstring
        n = ng_minimum(error=dataset.error, ng_max=self.nmax)
        quad_type = dataset.ng.attrs.get("quadrature_type", "gauss_legendre")
        return Quad.new(type=quad_type, n=n)


@parse_docs
@attrs.define
class QuadSpecErrorThreshold(QuadSpec):
    """
    Error-threshold number of quadrature points [``error_threshold``]

    Find the number of quadrature points so that the error on the atmospheric
    transmittance is below a specified threshold.
    """

    threshold: float = documented(
        attrs.field(),
        doc="Error threshold value",
        type="float",
    )

    nmax: int | None = documented(
        attrs.field(
            default=None,
            converter=attrs.converters.optional(int),
        ),
        doc="Maximum number of quadrature points",
        type="int or None",
        init_type="int, optional",
    )

    @staticmethod
    def from_dict(value: dict[str, t.Any]) -> QuadSpecErrorThreshold:
        # Inherit docstring
        return QuadSpecErrorThreshold(**value)

    def make_quad(self, dataset: xr.Dataset) -> Quad:
        # Inherit docstring
        quad_type = dataset.ng.attrs.get("quadrature_type", "gauss_legendre")
        n = ng_threshold(
            error=dataset.error, threshold=self.threshold, ng_max=self.nmax
        )
        return Quad.new(type=quad_type, n=n)


def ng_minimum(error: xr.DataArray, ng_max: int | None = None):
    """
    Find the number of quadrature points that minimizes the error.

    Parameters
    ----------
    error : DataArray
        Error data.

    ng_max : int, optional
        Maximum number of quadrature points. If not provided, it will be
        inferred from the error data.

    Returns
    -------
    int
        Number of quadrature points that minimizes the error.
    """

    if ng_max is None:
        ng_max = int(error.ng.max())

    error_w0 = error.isel(w=0)
    ng_min = int(error.ng.where(error_w0 == error_w0.min(), drop=True)[0])
    return ng_max if ng_min > ng_max else ng_min


def ng_threshold(
    error: xr.DataArray,
    threshold: float,
    ng_max: int | None = None,
):
    """
    Find the number of quadrature points so that the error is (strictly) below
    a specified threshold value.

    Parameters
    ----------
    error : DataArray
        Error data.

    threshold : float
        Error threshold.

    ng_max : int, optional
        Maximum number of quadrature points. If not provided, it will be
        inferred from the error data.

    Returns
    -------
    int
        Number of quadrature points so that the error is below the threshold.
    """

    if ng_max is None:
        ng_max = int(error.ng.max())

    error_w0 = error.isel(w=0)
    ng = error.ng.where(error_w0 < threshold, drop=True)

    if ng.size == 0:
        return ng_max
    else:
        ng = int(ng[0])
        return ng_max if ng > ng_max else ng


# ------------------------------------------------------------------------------
#                              Bin set data class
# ------------------------------------------------------------------------------


@parse_docs
@attrs.define(eq=False, frozen=True, slots=True)
class BinSet(SpectralSet):
    """
    A data class representing a bin set used in CKD mode.

    See Also
    --------
    :class:`.WavelengthSet`
    """

    bins: list[Bin] = documented(
        attrs.field(
            converter=list,
            validator=attrs.validators.deep_iterable(
                member_validator=attrs.validators.instance_of(Bin)
            ),
        ),
        doc="Set of bins.",
        type="list of :class:`.Bin`",
        init_type="iterable of :class:`.Bin`",
    )

    def spectral_indices(self) -> t.Generator[CKDSpectralIndex]:
        for bin in self.bins:
            yield from bin.spectral_indices()

    @property
    def wavelengths(self) -> pint.Quantity:
        return self.wcenters

    @property
    def wcenters(self) -> pint.Quantity:
        """
        Return the central wavelength of all bins.
        """
        units = ucc.get("wavelength")
        return [bin.wcenter.m_as(units) for bin in self.bins] * units

    @property
    def wmins(self) -> pint.Quantity:
        """
        Return the lower bound of all bins.
        """
        units = ucc.get("wavelength")
        return [bin.wmin.m_as(units) for bin in self.bins] * units

    @property
    def wmaxs(self) -> pint.Quantity:
        """
        Return the upper bound of all bins.
        """
        units = ucc.get("wavelength")
        return [bin.wmax.m_as(units) for bin in self.bins] * units

    @classmethod
    @ureg.wraps(None, (None, "nm", "nm", "nm", None), strict=False)
    def arange(
        cls,
        start: pint.Quantity,
        stop: pint.Quantity,
        step: pint.Quantity = 10.0 * ureg.nm,
        quad: Quad | None = None,
    ) -> BinSet:
        """
        Generate a bin set with linearly spaced bins.

        Parameters
        ----------
        start : quantity or float
            Lower bound of first bin. If a float is passed, it is interpreted as
            being in units of nm.

        stop : quantity
            Upper bound of last bin. If a float is passed, it is interpreted as
            being in units of nm.

        step : quantity, default: 10 nm
            Bin width. If a float is passed, it is interpreted as being in units
            of nm.

        quad : .Quad, optional
            Quadrature rule (same for all bins in the set). Defaults to
            a one-point Gauss-Legendre quadrature.

        Returns
        -------
        :class:`.BinSet`
            Generated bin set.
        """
        if quad is None:
            quad = Quad.gauss_legendre(1)

        wmins = np.arange(start, stop, step)
        wmaxs = wmins + step

        wunits = ucc.get("wavelength")
        bins = [
            Bin(
                wmin=(wmin * ureg.nm).to(wunits),
                wmax=(wmax * ureg.nm).to(wunits),
                quad=quad,
            )
            for wmin, wmax in zip(wmins, wmaxs)
        ]

        return cls(bins)

    @classmethod
    def from_srf(
        cls,
        srf: xr.Dataset,
        step: pint.Quantity = 10.0 * ureg.nm,
        quad: Quad | None = None,
    ) -> BinSet:
        """
        Generate a bin set with linearly spaced bins covering the spectral
        range of a spectral response function.

        Parameters
        ----------
        srf : Dataset
            Spectral response function dataset.

        step : quantity
            Wavelength step.

        quad : .Quad, optional
            Quadrature rule (same for all bins in the set). Defaults to
            a one-point Gauss-Legendre quadrature.

        Returns
        -------
        :class:`.BinSet`
            Generated bin set.
        """
        wavelengths = to_quantity(srf.w)
        wmin = wavelengths.min()
        wmax = wavelengths.max()

        return cls.arange(start=wmin - step, stop=wmax + step, step=step, quad=quad)

    @classmethod
    def from_wavelength_bounds(
        cls, wmin: pint.Quantity, wmax: pint.Quantity, quad: Quad | None = None
    ) -> BinSet:
        if quad is None:
            quad = Quad.gauss_legendre(1)

        return cls(
            bins=[
                Bin(wmin=_wmin, wmax=_wmax, quad=quad)
                for _wmin, _wmax in zip(np.atleast_1d(wmin), np.atleast_1d(wmax))
            ]
        )

    @classmethod
    def from_absorption_database(
        cls,
        abs_db: CKDAbsorptionDatabase,
        quad_spec: QuadSpec | None = None,
    ) -> BinSet:
        """
        Generate a bin set from a molecular absorption database.

        Parameters
        ----------
        abs_db : .CKDAbsorptionDatabase
            Molecular absorption database.

        quad_spec : .QuadSpec
            Quadrature rule specification. If provided, it will be used to
            generate the quadrature rule based on error data in the
            absorption dataset.

        Returns
        -------
        .BinSet
            Generated bin set.
        """
        if quad_spec is None:
            quad_spec = QuadSpec.default()

        # Note: We had to disable the advanced quadrature rule specification
        # because it requires to open all files in the absorption database, which
        # results  in an impracticable performance penalty with very large CKD
        # databases.
        # This feature will be enabled again in the future, when the spectral
        # loop handling will be rewritten.
        if isinstance(quad_spec, (QuadSpecMinError, QuadSpecErrorThreshold)):
            n_g = quad_spec.nmax if quad_spec.nmax is not None else _DEFAULT_NG
            warnings.warn(
                "Advanced quadrature specification is temporarily disabled, "
                "switching to a fixed quadrature rule specification with %s g-points."
                % n_g
            )
            quad_spec = QuadSpecFixed(n=n_g)

        quad = quad_spec.make_quad(None)  # TODO: Refactor this, it's an ugly hack
        wmin = abs_db.spectral_coverage["wbound_lower [nm]"].values * ureg.nm
        wmax = abs_db.spectral_coverage["wbound_upper [nm]"].values * ureg.nm
        return cls.from_wavelength_bounds(wmin=wmin, wmax=wmax, quad=quad)

    @classmethod
    def default(cls):
        """
        Generate a default bin set, which covers Eradiate's default spectral
        range with 10 nm-wide bins.
        """
        wmin = round_to_multiple(SPECTRAL_RANGE_MIN.m_as(ureg.nm), 10.0, "nearest")
        wmax = round_to_multiple(SPECTRAL_RANGE_MAX.m_as(ureg.nm), 10.0, "nearest")
        dw = 10.0

        return BinSet.arange(
            start=wmin * ureg.nm, stop=wmax * ureg.nm, step=dw * ureg.nm
        )



from __future__ import annotations

import warnings

import mitsuba as mi
import numpy as np
import xarray as xr


def bitmap_to_dataarray(bmp: mi.Bitmap, dtype="float64") -> xr.DataArray:
    """
    Format Mitsuba bitmap data as an xarray data array.

    Parameters
    ----------
    bmp : mitsuba.core.Bitmap or ndarray
        Mitsuba bitmap to be converted to a dataset. A Numpy array can also be
        passed directly for compatibility (this feature is deprecated).

    dtype : dtype, optional
        Data type, forwarded to :func:`numpy.array`.

    Returns
    -------
    da : DataArray
        Bitmap data as an xarray array.

    Raises
    ------
    ValueError
        If `bmp` has an unsupported pixel format.
    """
    img = np.array(bmp, dtype=dtype)

    if isinstance(bmp, mi.Bitmap):
        try:
            pixel_formats = {
                mi.Bitmap.PixelFormat.Y: ["Y"],
                mi.Bitmap.PixelFormat.YA: ["Y", "A"],
                mi.Bitmap.PixelFormat.RGB: ["R", "G", "B"],
                mi.Bitmap.PixelFormat.RGBA: ["R", "G", "B", "A"],
                mi.Bitmap.PixelFormat.XYZ: ["X", "Y", "Z"],
                mi.Bitmap.PixelFormat.XYZA: ["X", "Y", "Z", "A"],
            }
            channels = pixel_formats[bmp.pixel_format()]
        except KeyError:
            raise ValueError(f"unsupported bitmap pixel format {bmp.pixel_format()}")

    else:
        warnings.warn(
            "Passing an array is deprecated (coordinate detection is limited)",
            DeprecationWarning,
        )
        if len(img.shape) < 3:
            channels = ["Y"]
        else:
            channels = {1: ["Y"], 3: ["R", "G", "B"]}[img.shape[2]]

    height = img.shape[0]
    width = img.shape[1]

    result = xr.DataArray(
        data=np.reshape(img, (height, width, -1)),
        dims=["y_index", "x_index", "channel"],
        coords={
            "y_index": (
                "y_index",
                range(height),
                {"long_name": "height pixel index"},
            ),
            "y": (
                "y_index",
                np.linspace(0, 1, height),
                {"long_name": "film height coordinate"},
            ),
            "x_index": (
                "x_index",
                range(width),
                {"long_name": "width pixel index"},
            ),
            "x": (
                "x_index",
                np.linspace(0, 1, width),
                {"long_name": "film width coordinate"},
            ),
            "channel": ("channel", channels, {"long_name": "film spectral channel"}),
        },
    )

    return result


def bitmap_to_dataset(bmp: mi.Bitmap, dtype="float64") -> xr.Dataset:
    """
    Format Mitsuba bitmap data as an xarray dataset.

    Parameters
    ----------
    bmp : mitsuba.core.Bitmap or ndarray
        Mitsuba bitmap to be converted to a dataset. A Numpy array can also be
        passed directly for compatibility (this feature is deprecated).

    dtype : dtype, optional
        Data type, forwarded to :func:`numpy.array`.

    Returns
    -------
    dataset : Dataset
        Bitmap data as an xarray dataset.

    See Also
    --------
    :func:`bitmap_to_dataarray`

    Notes
    -----
    This function exists for backward compatibility purposes.
    It calls :func:`bitmap_to_dataarray` and forwards its arguments to it, then
    wraps the generated data array in a dataset with an ``img`` data variable.
    """

    da = bitmap_to_dataarray(bmp, dtype)
    return xr.Dataset(data_vars={"img": da})



from __future__ import annotations

import attrs
import mitsuba as mi
import pint

from ._core import Atmosphere
from ..core import traverse
from ..phase import PhaseFunction, RayleighPhaseFunction, phase_function_factory
from ..spectra import AirScatteringCoefficientSpectrum, Spectrum, spectrum_factory
from ...attrs import documented, parse_docs
from ...contexts import KernelContext
from ...kernel import InitParameter, TypeIdLookupStrategy, UpdateParameter
from ...spectral.ckd import BinSet
from ...spectral.index import SpectralIndex
from ...spectral.mono import WavelengthSet
from ...units import unit_context_kernel as uck
from ...validators import has_quantity


@parse_docs
@attrs.define(eq=False, slots=False)
class HomogeneousAtmosphere(Atmosphere):
    """
    Homogeneous atmosphere scene element [``homogeneous``].

    This class builds an atmosphere consisting of a homogeneous medium with
    customizable collision coefficients and phase function.
    """

    sigma_s: Spectrum = documented(
        attrs.field(
            factory=AirScatteringCoefficientSpectrum,
            converter=spectrum_factory.converter("collision_coefficient"),
            validator=[
                attrs.validators.instance_of(Spectrum),
                has_quantity("collision_coefficient"),
            ],
        ),
        doc="Atmosphere scattering coefficient value.\n"
        "\n"
        "Can be initialised with a dictionary processed by "
        ":data:`~eradiate.scenes.spectra.spectrum_factory`.",
        type=":class:`~eradiate.scenes.spectra.Spectrum` or float",
        default=":class:`AirScatteringCoefficientSpectrum() "
        "<.AirScatteringCoefficientSpectrum>`",
    )

    sigma_a: Spectrum = documented(
        attrs.field(
            default=0.0,
            converter=spectrum_factory.converter("collision_coefficient"),
            validator=[
                attrs.validators.instance_of(Spectrum),
                has_quantity("collision_coefficient"),
            ],
        ),
        doc="Atmosphere absorption coefficient value. Defaults disable "
        "absorption.\n"
        "\n"
        "Can be initialised with a dictionary processed by "
        ":data:`~eradiate.scenes.spectra.spectrum_factory`.",
        type=":class:`~eradiate.scenes.spectra.Spectrum`",
        default="0.0 km**-1",
    )

    _phase: PhaseFunction = documented(
        attrs.field(
            factory=lambda: RayleighPhaseFunction(),
            converter=phase_function_factory.convert,
            validator=attrs.validators.instance_of(PhaseFunction),
        ),
        doc="Scattering phase function.\n"
        "\n"
        "Can be initialised with a dictionary processed by "
        ":data:`~eradiate.scenes.phase.phase_function_factory`.",
        type=":class:`~eradiate.scenes.phase.PhaseFunction`",
        default=":class:`RayleighPhaseFunction() <.RayleighPhaseFunction>`",
    )

    def __attrs_post_init__(self) -> None:
        self.update()

    def update(self) -> None:
        # Inherit docstring
        self.phase.id = self.phase_id

    def spectral_set(self, quad_spec: dict) -> None | BinSet | WavelengthSet:
        return None

    # --------------------------------------------------------------------------
    #                               Properties
    # --------------------------------------------------------------------------

    @property
    def bottom(self) -> pint.Quantity:
        # Inherit docstring
        return self._bottom

    @property
    def top(self) -> pint.Quantity:
        # Inherit docstring
        return self._top

    @property
    def phase(self) -> PhaseFunction:
        # Inherit docstring
        return self._phase

    # --------------------------------------------------------------------------
    #                           Evaluation methods
    # --------------------------------------------------------------------------

    def eval_mfp(self, ctx: KernelContext) -> pint.Quantity:
        # Inherit docstring
        return (
            1.0 / self.eval_sigma_s(ctx.si)
            if self.eval_sigma_s(ctx.si).m != 0.0
            else 1.0 / self.eval_sigma_a(ctx.si)
        )

    def eval_albedo(self, si: SpectralIndex) -> pint.Quantity:
        """
        Return albedo at given spectral index.

        Parameters
        ----------
        si : :class:`.SpectralIndex`
            Spectral index.

        Returns
        -------
        quantity
            Albedo.
        """
        return self.eval_sigma_s(si) / (self.eval_sigma_s(si) + self.eval_sigma_a(si))

    def eval_sigma_a(self, si: SpectralIndex) -> pint.Quantity:
        """
        Return absorption coefficient at given spectral index.

        Parameters
        ----------
        si : :class:`.SpectralIndex`
            Spectral index.

        Returns
        -------
        quantity
            Absorption coefficient.
        """
        return self.sigma_a.eval(si)

    def eval_sigma_s(self, si: SpectralIndex) -> pint.Quantity:
        """
        Return scattering coefficient at given spectral index.

        Parameters
        ----------
        si : :class:`.SpectralIndex`
            Spectral index.

        Returns
        -------
        quantity
            Scattering coefficient.
        """
        return self.sigma_s.eval(si)

    def eval_sigma_t(self, si: SpectralIndex) -> pint.Quantity:
        """
        Return extinction coefficient at given spectral index.

        Parameters
        ----------
        si : :class:`.SpectralIndex`
            Spectral index.

        Returns
        -------
        quantity
            Extinction coefficient.
        """
        return self.eval_sigma_a(si) + self.eval_sigma_s(si)

    # --------------------------------------------------------------------------
    #                       Kernel dictionary generation
    # --------------------------------------------------------------------------

    @property
    def _template_phase(self) -> dict:
        # Inherit docstring
        result, _ = traverse(self.phase)
        return result.data

    @property
    def _template_medium(self) -> dict:
        # Inherit docstring
        return {
            "type": "homogeneous",
            "sigma_t": InitParameter(
                lambda ctx: self.eval_sigma_t(ctx.si).m_as(
                    uck.get("collision_coefficient")
                ),
            ),
            "albedo": InitParameter(
                lambda ctx: self.eval_albedo(ctx.si).m_as(uck.get("albedo"))
            ),
            # Note: "phase" is deliberately unset, this is left to the
            # Atmosphere.template property
        }

    @property
    def _params_medium(self) -> dict[str, UpdateParameter]:
        # Inherit docstring
        return {
            # Note: "value" appears twice because the mi.Spectrum is
            # encapsulated in a mi.ConstVolume
            "sigma_t.value.value": UpdateParameter(
                lambda ctx: self.eval_sigma_t(ctx.si).m_as(
                    uck.get("collision_coefficient")
                ),
                UpdateParameter.Flags.SPECTRAL,
                lookup_strategy=TypeIdLookupStrategy(
                    node_type=mi.Medium,
                    node_id=self.medium_id,
                    parameter_relpath="sigma_t.value.value",
                ),
            ),
            "albedo.value.value": UpdateParameter(
                lambda ctx: self.eval_albedo(ctx.si).m_as(uck.get("albedo")),
                UpdateParameter.Flags.SPECTRAL,
                lookup_strategy=TypeIdLookupStrategy(
                    node_type=mi.Medium,
                    node_id=self.medium_id,
                    parameter_relpath="albedo.value.value",
                ),
            ),
        }

    @property
    def _params_phase(self) -> dict[str, UpdateParameter]:
        # Inherit docstring
        _, params = traverse(self.phase)
        return params.data



from __future__ import annotations

import os
import typing as t
from abc import ABC, abstractmethod
from pathlib import Path

import attrs
import matplotlib.pyplot as plt
import mitsuba as mi
import numpy as np
import xarray as xr

from .. import data
from ..attrs import documented, parse_docs
from ..exceptions import DataError
from ..typing import PathLike
from ..util.misc import summary_repr


def regression_test_plots(
    ref: np.typing.ArrayLike,
    result: np.typing.ArrayLike,
    vza: np.typing.ArrayLike,
    filename: PathLike,
    metric: tuple[str, float],
) -> None:
    """
    Create regression test report plots.

    Parameters
    ----------
    ref : array-like
        BRF values for the reference data

    result : array-like
        BRF values for the simulation result

    vza : array-like
        VZA values for plotting

    filename : path-like
        Path to the output file for the plot

    metric : tuple
        A tuple of the form (metric name, value) to be added to the plots.
    """
    fig, axes = plt.subplots(2, 2, figsize=(8, 6))

    axes[0][0].plot(vza, ref, label="reference")
    axes[0][0].plot(vza, result, label="result")
    axes[0][0].set_title("Reference and test result")
    handles, labels = axes[0][0].get_legend_handles_labels()

    axes[1][0].plot(vza, result - ref)
    axes[1][0].set_xlabel("VZA [deg]")
    axes[1][0].set_ylabel("BRF in principal plane [-]")
    axes[1][0].set_title("Absolute difference")

    axes[1][1].plot(vza, (result - ref) / ref)
    axes[1][1].set_title("Relative difference")

    axes[0][1].set_axis_off()
    axes[0][1].legend(handles=handles, labels=labels, loc="upper center")
    axes[0][1].text(
        0.5, 0.5, f"{metric[0]}: {metric[1]:.4}", horizontalalignment="center"
    )

    plt.tight_layout()
    plt.savefig(filename)
    plt.close()


def reference_converter(
    value: os.PathLike | xr.Dataset | None,
) -> xr.Dataset | None:
    """
    A converter for handling the reference data attribute.

    Parameters
    ----------
    value : path-like or Dataset or None
        Path to the reference dataset file or dataset identifier or Dataset or
        None.

    Raises
    ------
    ValueError
        If the reference data is not a valid dataset.

    Returns
    -------
    xr.Dataset or None
        The reference dataset.

    Notes
    -----
    If value is ``None``, the converter returns ``None``.
    If value is a path to a local file, load it as a dataset.
    If value is a path to a remote file, load it from the Eradiate data store.
    If value is a dataset, return it as is.
    If value is a path to a remote file but the data store raised a DataError,
    returns ``None``.
    """
    if value is None:
        return value

    try:
        if isinstance(value, (str, os.PathLike, bytes)):
            # Try to open a file if it is directly referenced
            if os.path.isfile(value):
                return xr.load_dataset(value)

            # Try to serve the file from the data store
            return data.load_dataset(value)

        elif isinstance(value, xr.Dataset):
            return value

        else:
            raise ValueError(
                "Reference must be provided as a Dataset or a file path. "
                f"Got {type(value).__name__}"
            )

    except DataError:
        return None


@parse_docs
@attrs.define
class RegressionTest(ABC):
    """
    Common interface for tests based on the comparison of a result array against
    reference values.
    """

    # Name used for the reference metric. Must be set be subclasses.
    METRIC_NAME: t.ClassVar[str | None] = None

    name: str = documented(
        attrs.field(validator=attrs.validators.instance_of(str)),
        doc="Test case name.",
        type="str",
        init_type="str",
    )

    value: xr.Dataset = documented(
        attrs.field(
            validator=attrs.validators.instance_of(xr.Dataset),
            repr=summary_repr,
        ),
        doc="Simulation result. Must be specified as a dataset.",
        type=":class:`xarray.Dataset`",
        init_type=":class:`xarray.Dataset`",
    )

    reference: xr.Dataset | None = documented(
        attrs.field(
            default=None,
            converter=reference_converter,
            validator=attrs.validators.optional(
                attrs.validators.instance_of(xr.Dataset)
            ),
        ),
        doc="Reference data. Can be specified as an xarray dataset, a path to a "
        "NetCDF file or a path to a resource served by the data store.",
        type=":class:`xarray.Dataset` or None",
        init_type=":class:`xarray.Dataset` or path-like, optional",
        default="None",
    )

    threshold: float = documented(
        attrs.field(kw_only=True),
        doc="Threshold for test evaluation",
        type="float",
        init_type="float",
    )

    archive_dir: Path = documented(
        attrs.field(kw_only=True, converter=Path),
        doc="Path to output artefact storage directory. Relative paths are "
        "interpreted with respect to the current working directory.",
        type=":class:`pathlib.Path`",
        init_type="path-like",
    )

    def __attrs_pre_init(self):
        if self.METRIC_NAME is None:
            raise TypeError(f"Unsupported test type {type(self).__name__}")

    def run(self) -> bool:
        """
        This method controls the execution steps of the regression test:

        * handle missing reference data
        * catch errors during text evaluation
        * create the appropriate plots and data archives

        Returns
        -------
        bool
            Result of the test criterion comparison.
        """
        fname = self.name
        ext = ".nc"
        archive_dir = os.path.abspath(self.archive_dir)

        # Absolute path where the reference will be stored
        fname_reference = os.path.join(archive_dir, fname + "-ref" + ext)
        # Absolute path where test output will be stored
        fname_result = os.path.join(archive_dir, fname + "-result" + ext)

        # if no valid reference is found, store the results as new ref and fail
        # the test
        if not self.reference:
            print("No reference data was found! Storing test results as reference.")
            self._archive(self.value, fname_reference)
            self._plot(reference_only=True, metric_value=None)
            return False

        # else (we have a reference value), evaluate the test metric
        try:
            passed, metric_value = self._evaluate()
        except Exception as e:
            print("An exception occurred during test execution!")
            self._plot(reference_only=False, metric_value=None)
            raise e

        # we got a metric: report the results in the archive directory
        self._archive(self.value, fname_result)
        self._archive(self.reference, fname_reference)
        self._plot(reference_only=False, metric_value=metric_value)

        return passed

    @abstractmethod
    def _evaluate(self) -> tuple[bool, float]:
        """
        Evaluate the test results and perform a comparison to the reference
        based on the criterion defined in the specialized class.

        Returns
        -------
        passed : bool
            ``True`` iff the test passed.

        metric_value : float
            The value of the test metric.
        """
        pass

    def _archive(self, dataset: xr.Dataset, fname_output: PathLike) -> None:
        """
        Create an archive file for test result and reference storage
        """
        os.makedirs(os.path.dirname(fname_output), exist_ok=True)
        print(f"Saving dataset to {fname_output}")
        dataset.to_netcdf(fname_output)

    def _plot(self, metric_value: float | None, reference_only: bool) -> None:
        """
        Create a plot to visualize the results of the test.
        If the ``reference only`` parameter is set, create only a simple plot
        visualizing the new reference data. Otherwise, create the more complex
        comparsion plots for the regression test.

        Parameters
        ----------
        metric_value : float or None
            The numerical value of the test metric.

        reference_only : bool
            If ``True``, create only a simple visualization of the computed
            data.
        """
        vza = np.squeeze(self.value.vza.values)

        if "brf_srf" in self.value.data_vars:  # Handle spectral results
            brf = np.squeeze(self.value.brf_srf.values)
        else:  # Handle monochromatic results
            brf = np.squeeze(self.value.brf.values)

        fname = self.name
        ext = ".png"
        archive_dir = os.path.abspath(self.archive_dir)
        fname_plot = os.path.join(archive_dir, fname + ext)
        os.makedirs(os.path.dirname(fname_plot), exist_ok=True)
        print(f"Saving plot to {fname_plot}")

        if reference_only:
            plt.figure(figsize=(8, 6))
            plt.plot(vza, brf)
            plt.xlabel("VZA [deg]")
            plt.ylabel("BRF in principal plane [-]")
            plt.title("Simulation result, can be used as new reference")
            plt.tight_layout()
            plt.savefig(fname_plot)
            plt.close()

        else:
            if "brf_srf" in self.value.data_vars:  # Handle spectral results
                brf_ref = np.squeeze(self.reference.brf_srf.values)
            else:  # Handle monochromatic results
                brf_ref = np.squeeze(self.reference.brf.values)

            regression_test_plots(
                brf_ref, brf, vza, fname_plot, (self.METRIC_NAME, metric_value)
            )


@parse_docs
@attrs.define
class RMSETest(RegressionTest):
    """
    This class implements a simple test based on the root mean squared
    error (RMSE) of a result array with respect to the reference data.

    The test will pass if the computed root mean squared deviation between
    the result and reference is smaller or equal to the given threshold.
    """

    METRIC_NAME = "rmse"

    def _evaluate(self) -> tuple[bool, float]:
        value_np = self.value.brf.values
        ref_np = self.reference.brf.values
        if np.shape(value_np) != np.shape(ref_np):
            raise ValueError(
                f"Result and reference do not have the same shape! "
                f"Got: {np.shape(value_np)}, {np.shape(ref_np)}"
            )

        result_flat = np.array(value_np).flatten()
        ref_flat = np.array(ref_np).flatten()

        rmse = np.linalg.norm(result_flat - ref_flat) / np.sqrt(len(ref_flat))
        return rmse <= self.threshold, rmse


@parse_docs
@attrs.define
class Chi2Test(RegressionTest):
    """
    This class implements a statistical test for the regression testing
    campaign, based on Pearson's Chi-squared test.
    https://en.wikipedia.org/wiki/Pearson%27s_chi-squared_test

    It determines the probability for the reference and the test result
    following the same distribution.

    This test will pass if the computed p-value is strictly larger than the
    given threshold.
    """

    # The algorithm is adapted from Mitsuba's testing framework.

    METRIC_NAME = "p-value"

    def _evaluate(self) -> tuple[bool, float]:
        ref_np = self.reference.brf.values

        result_np = self.value.brf.values
        histo_bins = np.linspace(ref_np.min(), ref_np.max(), 20)
        histo_ref = np.histogram(ref_np, histo_bins)[0]
        histo_res = np.histogram(result_np, histo_bins)[0]

        # sorting both histograms following the ascending frequencies in
        # the reference. Algorithm from:
        # https://stackoverflow.com/questions/9764298/how-to-sort-two-lists-which-reference-each-other-in-the-exact-same-way
        histo_ref_sorted, histo_res_sorted = zip(
            *sorted(zip(histo_ref, histo_res), key=lambda x: x[0])
        )

        from mitsuba.python.math import rlgamma

        chi2val, dof, pooled_in, pooled_out = mi.math.chi2(
            histo_res_sorted, histo_ref_sorted, 5
        )
        p_value = 1.0 - rlgamma(dof / 2.0, chi2val / 2.0)

        return p_value > self.threshold, p_value



from __future__ import annotations

import warnings

import attrs
import mitsuba as mi
import numpy as np
import pint
import pinttr

import eradiate

from ._core import Surface
from ..bsdfs import BSDF, BlackBSDF, LambertianBSDF, bsdf_factory
from ..core import Ref, SceneTraversal, traverse
from ..shapes import RectangleShape, shape_factory
from ...attrs import documented, parse_docs
from ...exceptions import OverriddenValueWarning, TraversalError
from ...units import unit_context_config as ucc


def _edges_converter(value):
    # Basic unit conversion and array reshaping
    length_units = ucc.get("length")
    value = np.reshape(
        pinttr.util.ensure_units(value, default_units=length_units).m_as(length_units),
        (-1,),
    )

    # Broadcast if relevant
    if len(value) == 1:
        value = np.full((2,), value[0])

    return value * length_units


@parse_docs
@attrs.define(eq=False, slots=False)
class CentralPatchSurface(Surface):
    """
    Central patch surface [``central_patch``].

    This surface consists of a rectangular patch, described by its `field`
    parameter, with a composite reflection model composed of a background
    uniform component, and a central patch.


    This class creates a square surface to which two BSDFs will be attached.

    The two constituent surfaces ``central_patch`` and ``background_surface``
    define the properties of the two sections of this surface.

    The size of the central surface is controlled by setting the ``width``
    parameter of the ``central_patch`` surface, while the ``width`` of the
    ``background_surface`` must be set to ``AUTO`` and the total width of the
    surface is set by the ``width`` of the main surface object.
    Note that the ``width`` of a surface defaults to ``AUTO``, which means,
    omitting the parameter in the ``background_surface`` will yield the correct
    behaviour.

    If the ``central_patch`` width is set to ``AUTO`` as well it defaults to one
    third of the overall surface size, unless a contextual constraint (*e.g.* to
    match the size of an atmosphere or canopy) is applied.
    """

    shape: RectangleShape | None = documented(
        attrs.field(
            default=None,
            converter=attrs.converters.optional(shape_factory.convert),
            validator=attrs.validators.optional(
                attrs.validators.instance_of(RectangleShape)
            ),
        ),
        doc="Shape describing the surface. This parameter may be left unset "
        "for situations in which setting its value is delegated to another "
        "component (*e.g.* an :class:`.Experiment` instance owning the "
        "surface object); however, if it is still unset upon kernel dictionary "
        "generation, the call to :meth:`.kernel_dict` will raise.",
        type=".RectangleShape or None",
        init_type=".RectangleShape or dict, optional",
        default="None",
    )

    @shape.validator
    def _shape_validator(self, attribute, value):
        if value is not None:  # Means it's a Shape
            if value.bsdf is not None:
                warnings.warn(
                    f"while validating '{attribute.name}': "
                    f"'{attribute.name}.bsdf' should be set to None; it will "
                    "be overridden upon kernel dictionary generation",
                    OverriddenValueWarning,
                )

    bsdf: BSDF = documented(
        attrs.field(
            factory=LambertianBSDF,
            converter=bsdf_factory.convert,
            validator=attrs.validators.instance_of(BSDF),
        ),
        doc="The reflection model attached to the surface.",
        type=".BSDF",
        init_type=".BSDF or dict, optional",
        default=":class:`LambertianBSDF() <.LambertianBSDF>`",
    )

    patch_edges: pint.Quantity | None = documented(
        pinttr.field(
            default=None,
            converter=attrs.converters.optional(_edges_converter),
            units=ucc.deferred("length"),
        ),
        doc="Length of the central patch's edges. If unset, the central patch "
        "edges will be 1/3 of the surface's edges. "
        "Unit-enabled field (default: ``ucc['length']``).",
        type="quantity or None",
        init_type="quantity or array-like, optional",
    )

    patch_bsdf: BSDF = documented(
        attrs.field(
            factory=BlackBSDF,
            converter=bsdf_factory.convert,
            validator=attrs.validators.instance_of(BSDF),
        ),
        doc="The reflection model attached to the central patch.",
        type=".BSDF",
        init_type=".BSDF or dict, optional",
        default=":class:`BlackBSDF() <.BlackBSDF>`",
    )

    def _texture_scale(self):
        """
        Compute patch texture scaling factor based on configuration.
        """
        # Note: The texture file has a central patch covering 1/3 of its
        # surface, hence the 1/3 factor.
        return (
            [1.0, 1.0]
            if self.patch_edges is None
            else (self.shape.edges / (3.0 * self.patch_edges)).m_as("dimensionless")
        )

    def update(self) -> None:
        # Inherit docstring

        # Fix BSDF IDs
        self.bsdf.id = self._background_bsdf_id
        self.patch_bsdf.id = self._patch_bsdf_id

        # Force BSDF referencing if the shape is defined
        if self.shape is not None:
            if isinstance(self.shape.bsdf, BSDF):
                warnings.warn("Set BSDF will be overridden by surface BSDF settings.")
            self.shape.bsdf = Ref(id=self._bsdf_id)

    @property
    def _shape_id(self):
        """
        Mitsuba shape object identifier.
        """
        return f"{self.id}_shape"

    @property
    def _bsdf_id(self):
        """
        Mitsuba BSDF object identifier (blend).
        """
        return f"{self.id}_bsdf"

    @property
    def _background_bsdf_id(self):
        """
        Mitsuba BSDF object identifier (background).
        """
        return f"{self.id}_background_bsdf"

    @property
    def _patch_bsdf_id(self):
        """
        Mitsuba BSDF object identifier (patch).
        """
        return f"{self.id}_patch_bsdf"

    @property
    def _template_bsdfs(self) -> dict:
        objects = {
            "bsdf_0": traverse(self.bsdf)[0].data,
            "bsdf_1": traverse(self.patch_bsdf)[0].data,
        }

        scale = self._texture_scale()
        to_uv = mi.ScalarTransform4f.scale(
            [scale[0], scale[1], 1]
        ) @ mi.ScalarTransform4f.translate(
            [-0.5 + (0.5 / scale[0]), -0.5 + (0.5 / scale[1]), 0.0]
        )

        result = {f"{self._bsdf_id}.type": "blendbsdf"}

        for obj_key, obj_params in objects.items():
            for key, param in obj_params.items():
                result[f"{self._bsdf_id}.{obj_key}.{key}"] = param

        weight_dict = {
            "type": "bitmap",
            "filename": str(
                eradiate.data.data_store.fetch(
                    "textures/central_patch_surface_mask.bmp"
                )
            ),
            "filter_type": "nearest",
            "to_uv": to_uv,
            "wrap_mode": "clamp",
        }

        for key, param in weight_dict.items():
            result[f"{self._bsdf_id}.weight.{key}"] = param

        return result

    @property
    def _template_shapes(self) -> dict:
        kdict_template = traverse(self.shape)[0].data

        result = {}

        for key, param in kdict_template.items():
            result[f"{self._shape_id}.{key}"] = param

        return result

    @property
    def _params_bsdfs(self) -> dict:
        objects = {
            "bsdf_0": traverse(self.bsdf)[1].data,
            "bsdf_1": traverse(self.patch_bsdf)[1].data,
        }

        result = {}

        for obj_key, obj_params in objects.items():
            for key, param in obj_params.items():
                result[f"{self._bsdf_id}.{obj_key}.{key}"] = param

        return result

    @property
    def _params_shapes(self) -> dict:
        return {}

    def traverse(self, callback: SceneTraversal) -> None:
        # Inherit docstring

        if self.shape is None:
            raise TraversalError(
                "A 'CentralPatchSurface' cannot be traversed if its 'shape' field "
                "is unset."
            )

        super().traverse(callback)



"""
System information display. This script is useful to collect environment
information. It is the fundamental building block of the `eradiate show`
command.

This code is partly taken from `mitsuba.sys_info`.
"""

import importlib
import locale
import platform
import re
import subprocess
import sys
from importlib import import_module
from importlib.metadata import PackageNotFoundError, version

import drjit as dr
import mitsuba as mi


def _run(command):
    """Returns (return-code, stdout, stderr)"""
    p = subprocess.Popen(
        command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True
    )
    raw_output, raw_err = p.communicate()
    rc = p.returncode
    if sys.platform.startswith("win32"):
        enc = "oem"
    else:
        enc = locale.getpreferredencoding()
    output = raw_output.decode(enc)
    err = raw_err.decode(enc)
    return rc, output.strip(), err.strip()


def _run_and_match(command, regex):
    """Runs command and returns the first regex match if it exists"""
    rc, out, _ = _run(command)
    if rc != 0:
        return None
    match = re.search(regex, out)
    if match is None:
        return None
    return match.group(1)


def _get_cpu_info():
    if sys.platform.startswith("win32"):
        return platform.processor()
    elif sys.platform.startswith("darwin"):
        return subprocess.check_output(
            ["/usr/sbin/sysctl", "-n", "machdep.cpu.brand_string"]
        ).strip()
    elif sys.platform.startswith("linux"):
        command = "cat /proc/cpuinfo"
        all_info = subprocess.check_output(command, shell=True).decode().strip()
        for line in all_info.split("\n"):
            if "model name" in line:
                return re.sub(r".*model name.*:", "", line, 1)[1:]
    return None


def show() -> dict:
    mi.set_variant("scalar_rgb")
    result = {}

    if sys.platform.startswith("darwin"):
        result["os"] = _run_and_match("sw_vers -productVersion", r"(.*)")
    elif sys.platform.startswith("win32"):
        result["os"] = platform.platform(terse=True)
    else:
        result["os"] = _run_and_match("lsb_release -a", r"Description:\t(.*)")

    result["cpu_info"] = _get_cpu_info()
    result["python"] = sys.version.replace("\n", "")
    result["llvm_version"] = dr.llvm_version()
    result["drjit_version"] = dr.__version__ + (" (DEBUG)" if dr.DEBUG else "")
    result["mitsuba_version"] = mi.MI_VERSION + (" (DEBUG)" if mi.DEBUG else "")
    try:
        eradiate_mitsuba_version = version("eradiate-mitsuba")
    except PackageNotFoundError:
        eradiate_mitsuba_version = "not installed [DEV|DEBUG]"
    result["eradiate_mitsuba_version"] = eradiate_mitsuba_version
    result["mitsuba_compiler"] = import_module("mitsuba.config").CXX_COMPILER

    # Python dependencies
    for package in ["numpy", "scipy", "xarray"]:
        try:
            mod = importlib.import_module(package)
            result[package] = mod.__version__
        except PackageNotFoundError:
            result[package] = "not found"

    return result


if __name__ == "__main__":
    sys_info = show()

    print("System")
    print(f"  CPU: {sys_info['cpu_info']}")
    print(f"  OS: {sys_info['os']}")
    print(f"  Python: {sys_info['python']}")

    print("\nVersions")
    print(f"  llvm {sys_info['llvm_version']}")
    print(f"  drjit {sys_info['drjit_version']}")
    print(
        f"  eradiate-mitsuba {sys_info['eradiate_mitsuba_version']} "
        f"(based on mitsuba {sys_info['mitsuba_version']})"
    )

    print("\nMitsuba variants")
    print("\n".join([f"  {variant}" for variant in mi.variants()]))



"""
Volume data file I/O helpers.
"""
from __future__ import annotations

import mitsuba as mi
import numpy as np
import xarray as xr

from ..typing import PathLike


def write_binary_grid3d(filename: PathLike, values: np.ndarray | xr.DataArray) -> None:
    """
    Write volume data to a binary file so that a ``gridvolume`` kernel plugin
    can be instantiated with that file.

    Parameters
    ----------
    filename : path-like
        File name.

    values : ndarray or DataArray
        Data array to output to the volume data file. This array must have 3 or
        4 dimensions (x, y, z, spectrum). If the array is 3-dimensional, it will
        automatically be assumed to have only one spectral channel.
    """
    if isinstance(values, xr.DataArray):
        values = values.values

    if not isinstance(values, np.ndarray):
        raise TypeError(
            f"unsupported data type {type(values)} "
            f"(expected numpy array or xarray DataArray)"
        )

    if values.ndim not in {3, 4}:
        raise ValueError(
            f"'values' must have 3 or 4 dimensions " f"(got shape {values.shape})"
        )

    mi.VolumeGrid(values.astype(np.float32)).write(str(filename))


def read_binary_grid3d(filename: PathLike) -> np.ndarray:
    """
    Reads a volume data binary file.

    Parameters
    ----------
    filename : path-like
        File name.

    Returns
    -------
    ndarray
        Values.
    """
    return np.array(mi.VolumeGrid(str(filename)))



from __future__ import annotations

import logging
import os
import shutil
import time
import typing as t
from pathlib import Path

import attrs
import pooch
from requests import RequestException

from ._core import DataStore, expand_rules
from ..attrs import documented, parse_docs
from ..exceptions import DataError
from ..typing import PathLike
from ..util.misc import LoggingContext

logger = logging.getLogger(__name__)


@parse_docs
@attrs.define
class BlindOnlineDataStore(DataStore):
    """
    Serve data downloaded from a remote source without integrity check.
    """

    _base_url: str = documented(
        attrs.field(converter=lambda x: x + "/" if not x.endswith("/") else x),
        type="str",
        doc="URL to the online storage location.",
    )

    path: Path = documented(
        attrs.field(converter=lambda x: Path(x).absolute()),
        type="Path",
        init_type="path-like",
        doc="Path to the local cache location.",
    )

    attempts: int = documented(
        attrs.field(default=3, converter=int),
        type="int",
        doc="Number of download attempts to make before giving up because of "
        "connection errors.",
    )

    @property
    def base_url(self) -> str:
        # Inherit docstring
        return self._base_url

    @property
    def registry(self) -> dict:
        """
        Raises :class:`NotImplementedError` (this data store has no registry).
        """
        raise NotImplementedError

    def registry_files(
        self, filter: t.Callable[[t.Any], bool] | None = None
    ) -> list[str]:
        """
        Returns an empty list (this data store has no registry).
        """
        return []

    def fetch(self, filename: PathLike, downloader: t.Callable | None = None) -> Path:
        """
        Fetch a file from the data store. This method wraps
        :func:`pooch.retrieve` and automatically selects compressed files
        when they are available.

        Parameters
        ----------
        filename : path-like
            File name to fetch from the local storage, relative to the storage
            root.

        downloader : callable, optional
            A callable that will be called to download a given URL to a provided
            local file name. This is mostly useful to
            `display progress bars <https://www.fatiando.org/pooch/latest/progressbars.html>`_
            during download.

        Returns
        -------
        path : Path
            Absolute path where the retrieved resource is located.

        Notes
        -----
        If a compressed resource exists, it will be served automatically.
        For instance, if ``"foo.nc"`` is requested and ``"foo.nc.gz"`` is
        registered, the latter will be downloaded, decompressed and served as
        ``"foo.nc"``.
        """

        fname = Path(filename).as_posix()
        url = self.base_url + fname
        max_wait = 10
        result = None

        with LoggingContext(
            pooch.get_logger(), level="WARNING"
        ):  # Silence pooch messages temporarily
            for i in range(self.attempts):
                # Try first to get a compressed file
                try:
                    result = Path(
                        pooch.retrieve(
                            url + ".gz",
                            known_hash=None,
                            fname=fname + ".gz",
                            path=self.path,
                            processor=pooch.processors.Decompress(
                                name=os.path.basename(fname)
                            ),
                        )
                    )
                    break
                except RequestException:
                    pass

                # If no gzip-compressed file is available, try the actual file
                try:
                    result = Path(
                        pooch.retrieve(
                            url,
                            known_hash=None,
                            fname=fname,
                            path=self.path,
                        ),
                    )
                    break
                except RequestException:
                    pass

                # If we get here, it means download failed
                retries_left = self.attempts - (i + 1)
                logger.info(
                    "Failed to download '%s'. "
                    "Will attempt the download again %d more time%s.",
                    fname,
                    retries_left,
                    "s" if retries_left > 1 else "",
                )
                time.sleep(min(i + 1, max_wait))

        if result is None:
            raise DataError(
                f"file '{fname}' could not be retrieved from {self.base_url}"
            )
        else:
            return result

    def purge(self, keep: None | str | list[str] = None) -> None:
        """
        Purge local storage location. The default behaviour is very aggressive
        and will wipe out the entire directory contents.

        Parameters
        ----------
        keep : str or list of str, optional
            A list of exclusion rules (paths relative to the store's local
            storage root, shell wildcards allowed) defining files which should
            be excluded from the purge process.

        Warnings
        --------
        This is a destructive operation, make sure you know what you're doing!
        """
        if keep is not None and not isinstance(keep, (list, tuple)):
            keep = list(keep)

        # Fast track for simple case
        if not keep:
            for filename in os.scandir(self.path):
                file_path = os.path.join(self.path, filename)

                if os.path.isfile(file_path) or os.path.islink(file_path):
                    os.remove(file_path)
                elif os.path.isdir(file_path):
                    shutil.rmtree(file_path)

            return

        # List files to keep and delete
        included = expand_rules(rules=["**/*"], prefix=self.path)
        excluded = expand_rules(rules=keep, prefix=self.path)
        remove = sorted(included - excluded)

        for file in remove:
            os.remove(file)

        # Clean up empty directories
        for x in self.path.iterdir():
            if x.is_dir():
                try:
                    os.removedirs(x)
                except OSError:
                    pass



from __future__ import annotations

import os
import typing as t
import warnings
from abc import ABC, abstractmethod
from pathlib import Path

import attrs
import pooch
import tqdm
from ruamel.yaml import YAML

from ..typing import PathLike


@attrs.define
class DataStore(ABC):
    """
    Interface class for all data stores.
    """

    @property
    @abstractmethod
    def base_url(self) -> str:
        """
        str : Address of the remote storage location.
        """
        pass

    @property
    @abstractmethod
    def registry(self) -> dict:
        """
        dict : Registry contents.
        """
        pass

    @abstractmethod
    def registry_files(
        self, filter: t.Callable[[t.Any], bool] | None = None
    ) -> list[str]:
        """
        Get a list of registered files.

        Parameters
        ----------
        filter : callable, optional
            A filter function taking a file path as a single string argument and
            returning a boolean. Filenames for which the filter returns ``True``
            will be returned.

        Returns
        -------
        files : list of str
            List of registered files.
        """
        pass

    @abstractmethod
    def fetch(
        self,
        filename: PathLike,
        **kwargs,
    ) -> Path:
        """
        Fetch a file from the data store.

        Parameters
        ----------
        filename : path-like
            File name to fetch from the data store, relative to the storage root.

        Returns
        -------
        Path
            Absolute path where the retrieved resource is located.

        Raises
        ------
        DataError
            The requested file could not be served.
        """
        pass


def registry_from_file(filename: PathLike, warn: bool = True) -> dict:
    """
    Read registry content from a file.

    Parameters
    ----------
    filename : path-like
        Path to the file to be read.

    warn : bool, optional
        If ``True``, ill-formed lines will result in a warning. Otherwise, an
        exception will be raised.

    Returns
    -------
    registry : dict
        Registry contents, parsed as a dictionary.

    Raises
    ------
    ValueError
        If `warn` is ``False`` and an error occurs while parsing the file.
    """
    result = {}

    with open(filename, "r") as f:
        for i, line in enumerate(f):
            line = line.strip()

            # Skip empty lines
            if not line:
                continue

            # Skip comment lines
            if line.startswith("#"):
                continue

            items = line.split(maxsplit=1)

            # No metadata (just the filename)
            if len(items) == 1:
                result[items[0]] = ""
                continue

            # Found metadata
            if len(items) == 2:
                result[items[0]] = items[1]
                continue

            # Ill-formed line
            if len(items) > 2:
                if warn:
                    warnings.warn(
                        f"While parsing registry file {filename}: skipping "
                        f"ill-formed line {i}"
                    )
                    continue
                else:
                    raise ValueError(
                        f"While parsing registry file {filename}: "
                        f"ill-formed line {i}"
                    )

    return result


def registry_to_file(registry: dict, filename: PathLike) -> None:
    """
    Write a registry dictionary to a text file.

    Parameters
    ----------
    registry : dict
        A registry dictionary.

    filename : path-like
        Path to the file where registry contents are to be written.
    """
    lines = [f"{path} {registry[path]}".strip() for path in sorted(registry)]
    lines.append("")
    content = "\n".join(lines)
    with open(filename, "w") as f:
        f.write(content)


def load_rules(filename: PathLike) -> dict:
    """
    Load include and exclude rules from a YAML file.

    Parameters
    ----------
    filename : path-like
        Path to the YAML file from which rules are to be loaded.

    Returns
    -------
    rules : dict
        Dictionary containing a list of inclusion (resp. exclusion) rules under
        the ``"include"`` (resp. ``"exclude"``) key.
    """

    yaml = YAML()
    try:
        with open(filename) as f:
            rules = yaml.load(f)
    except OSError:
        rules = {"include": ["**/*.*"], "exclude": []}

    return rules


def expand_rules(
    rules: list[str],
    prefix: PathLike = ".",
    as_list: bool = False,
    include_dirs=False,
) -> list | set:
    """
    Expand a list of filesystem selection rules to paths.

    Parameters
    ----------
    rules : list of str
        List of inclusion to expand, relative to `prefix`. Each rule may be a
        path to a file or a shell glob.

    prefix : path-like, optional
        Path where to expand the rules. By default, the current working
        directory (``"."``) is used.

    as_list : bool, optional
        If ``True``, return the result as a sorted list; otherwise, return it
        as a set.

    include_dirs : bool, optional
        If ``True``, include directories in the list of expanded items.
        Otherwise, the expansion is restricted to files.

    Returns
    -------
    items : set or list
        Items corresponding to the expanded rules.
    """
    expanded = set()

    for rule in rules:
        expanded |= set(
            x for x in Path(prefix).rglob(rule) if (x.is_file() or include_dirs)
        )

    return sorted(expanded) if as_list else expanded


def list_files(
    path: PathLike,
    includes: list[str] | None = None,
    excludes: list[str] | None = None,
    as_list: bool = False,
) -> list | set:
    """
    List files in a directory based on inclusion and exclusion rules.

    Parameters
    ----------
    path : path-like
        Path to the target directory.

    includes : list of str, optional
        List of inclusion rules, relative to `path`. Each rule may be a path to
        a file or a shell glob. If no rule is passed, everything is included
        (*i.e.* ``["**/*"]`` is used).

    excludes : list of str, optional
        List of exclusion rules, relative to `path`. Each rule may be a path to
        a file or a shell glob. If no rule is passed, nothing is excluded
        (*i.e.* ``[]`` is used).

    as_list : bool
        If ``True``, return the result as a sorted list; otherwise, return it
        as a set.

    Returns
    -------
    items : set or list
        Files listed in the target directory, according to the rules.
    """
    if includes is None:
        includes = ["**/*"]
    if excludes is None:
        excludes = []

    included: set = expand_rules(includes, prefix=path)
    excluded: set = expand_rules(excludes, prefix=path)

    return sorted(included - excluded) if as_list else included - excluded


def make_registry(
    path: PathLike,
    filename: PathLike = "registry.txt",
    includes: list | None = None,
    excludes: list | None = None,
    alg="sha256",
    show_progress=False,
) -> None:
    """
    Create a registry file from items in a directory, possibly applying
    inclusion and exclusion rules.

    Parameters
    ----------
    path : path-like
        Path to the target directory.

    filename : path-like
        Path to the file where registry contents are to be written.

    includes : list of str, optional
        List of inclusion rules, relative to `path`. Each rule may be a path to
        a file or a shell glob. If no rule is passed, everything is included
        (*i.e.* ``["**/*"]`` is used).

    excludes : list of str, optional
        List of exclusion rules, relative to `path`. Each rule may be a path to
        a file or a shell glob. If no rule is passed, nothing is excluded
        (*i.e.* ``[]`` is used).

    alg : str, optional
        Hash algorithm used.
    """
    # Basically pooch.make_registry() with inclusion / exclusion rules
    files = list_files(path, includes, excludes)
    hashes = (
        [pooch.file_hash(x, alg=alg) for x in files]
        if alg is not None
        else ["" for _ in files]
    )

    registry = {}

    with tqdm.tqdm(total=len(hashes), disable=not show_progress) as pbar:
        for file, hash in zip(files, hashes):
            registry[os.path.relpath(file, path)] = f"{alg}:{hash}"
            pbar.update()

    registry_to_file(registry, filename)



"""
Manage files stored in the ``eradiate-data`` repository.
"""
from __future__ import annotations

import os
import typing as t
from pathlib import Path

import attrs

from ._core import DataStore, load_rules, make_registry, registry_from_file
from ..attrs import documented, parse_docs
from ..exceptions import DataError
from ..typing import PathLike


@parse_docs
@attrs.define
class SafeDirectoryDataStore(DataStore):
    """
    Serve files stored in a directory. This data store will only serve files
    listed in its registry.
    """

    path: Path = documented(
        attrs.field(converter=lambda x: Path(x).absolute()),
        type="Path",
        init_type="path-like",
        doc="Path to the root of the directory referenced by this data store.",
    )

    registry_fname: Path = documented(
        attrs.field(default="registry.txt", converter=Path),
        type="Path",
        init_type="path-like",
        default='"registry.txt"',
        doc="Path to the registry file, relative to `path`.",
    )

    @registry_fname.validator
    def _registry_fname_validator(self, attribute, value: Path):
        if value.is_absolute():
            raise ValueError(
                f"while validating '{attribute.name}': "
                "only paths relative to the store root path are allowed"
            )

    _registry: dict = attrs.field(factory=dict, converter=dict, repr=False, init=False)

    def __attrs_post_init__(self):
        self.registry_reload()

    @property
    def base_url(self) -> str:
        # Inherit docstring
        return str(self.path)

    @property
    def registry(self) -> dict:
        # Inherit docstring
        return self._registry

    def registry_files(
        self, filter: t.Callable[[t.Any], bool] | None = None
    ) -> list[str]:
        # Inherit docstring
        raise NotImplementedError

    @property
    def registry_path(self) -> Path:
        """
        Path: Path to the registry file.
        """
        return self.path / self.registry_fname

    def registry_make(self) -> None:
        """
        Generate a registry file from the contents of the ``self.path``
        directory, according to inclusion and exclusion rules defined in the
        ``self.path / "registry_rules.yml"`` file. The generated registry is
        written to ``self.path / self.registry_fname``.
        """
        # Load include and exclude rules
        rules = load_rules(self.path / "registry_rules.yml")

        # Write registry
        make_registry(
            self.path,
            self.registry_path,
            includes=rules["include"],
            excludes=rules["exclude"],
        )

    def registry_fetch(self) -> Path:
        """
        Get the absolute path to the registry file.
        If no file exists, one will be created based on the rules contained
        defined in ``self.path / "registry_rules.yml"``.
        """
        filename = self.registry_path

        if not filename.is_file():
            self.registry_make()

        return filename

    def registry_delete(self):
        """
        Delete the registry file.
        """
        os.remove(self.path / self.registry_fname)

    def registry_reload(self) -> None:
        """
        Reload the registry file from the hard drive.
        """
        self._registry = registry_from_file(self.registry_fetch())

    def fetch(self, filename: PathLike, **kwargs) -> Path:
        # No kwargs are actually accepted
        if kwargs:
            keyword = next(iter(kwargs.keys()))
            raise TypeError(f"fetch() got an unexpected keyword argument '{keyword}'")

        fname = Path(filename).as_posix()
        if fname in self.registry:
            return self.path / fname
        else:
            raise DataError(f"file '{fname}' is not in the registry")



from __future__ import annotations

import mitsuba as mi
import pytest

from ..contexts import KernelContext
from ..kernel import MitsubaObjectWrapper, mi_traverse
from ..scenes.core import CompositeSceneElement, NodeSceneElement, Scene, traverse


def check_scene_element(
    instance: NodeSceneElement | CompositeSceneElement,
    mi_cls=None,
    ctx: KernelContext = None,
    drop_parameters: bool = True,
) -> MitsubaObjectWrapper:
    """
    Perform kernel dictionary checks on a scene element.

    This function checks if the scene element can produce a valid kernel
    dictionary template, as well as an appropriate parameter table.

    The returned Mitsuba object and parameter table can be used to perform
    additional checks.

    Parameters
    ----------
    instance : :class:`.NodeSceneElement` or :class:`.CompositeSceneElement`
        Node scene element to check.

    mi_cls : :class:`mitsuba.Object`
        Mitsuba class the node scene element expands to. Must be set if
        `instance` is a :class:`.NodeSceneElement`; ignored otherwise.

    ctx : .KernelContext, optional
        If provided, the kernel dictionary context to use. Otherwise, a default
        context is created.

    drop_parameters : bool, default: True
        If ``True``, the Mitsuba scene parameter table will be stripped off from
        untracked parameters.

    Returns
    -------
    mi_obj : :class:`mitsuba.Object`
        Mitsuba object the scene element was expanded to. See notes for details.

    mi_params : dict
        Parameter table of the Mitsuba objects generated by the tested scene
        element.

    Notes
    -----
    * If `instance` is a :class:`.NodeSceneElement`, the corresponding Mitsuba
      object type is checked against `mi_cls`.
    * If `instance` is a :class:`.CompositeSceneElement`, the corresponding
      Mitsuba objects are automatically encapsulated into a
      :class:`mitsuba.Scene` object.
    """
    if isinstance(instance, NodeSceneElement):
        if mi_cls is None:
            raise ValueError("Expected Mitsuba class must be set")
        kdict_template, umap_template = traverse(instance)

    elif isinstance(instance, CompositeSceneElement):
        mi_cls = mi.Scene
        kdict_template, umap_template = traverse(Scene(objects={"composite": instance}))

    else:
        raise RuntimeError(f"Cannot test type '{instance.__class__}'")

    # Check if the template can be instantiated
    ctx = KernelContext() if ctx is None else ctx
    kernel_dict = kdict_template.render(ctx)

    try:
        mi_obj = mi.load_dict(kernel_dict)
    except RuntimeError as e:
        pytest.fail(
            reason=f"could not load scene dictionary, got RuntimeError: {e}\n"
            f"{kernel_dict = }"
        )

    assert isinstance(mi_obj, mi_cls)

    # Collect Mitsuba parameters, resolve update map parameter paths
    mi_wrapper = mi_traverse(mi_obj, umap_template)
    mi_params = mi_wrapper.parameters

    # Check that parameters can all be set
    umap = umap_template.render(ctx)
    for key, value in umap.items():
        try:
            mi_params[key] = value
        except KeyError as e:
            pytest.fail(
                reason=f"could not set parameter, got KeyError: {e}\n{mi_params = }"
            )

    mi_params.update()

    # Drop untracked parameters (this will detect param lookup failures)
    if drop_parameters:
        mi_wrapper.drop_parameters()

    return mi_wrapper



import logging
import os.path
import textwrap
from pathlib import Path
from typing import List, Optional

import typer
from rich.console import Console
from rich.logging import RichHandler
from ruamel.yaml import YAML
from typing_extensions import Annotated

import eradiate

from ..data._util import get_file_list, known_file_lists
from ..exceptions import DataError
from ..radprops._absorption import (
    KNOWN_DATABASES as KNOWN_MOLECULAR_ABSORPTION_DATABASES,
)
from ..radprops._absorption import AbsorptionDatabase

app = typer.Typer()
console = Console(color_system=None)

FORMAT = "%(message)s"
logging.basicConfig(
    level="INFO", format=FORMAT, datefmt="[%X]", handlers=[RichHandler()]
)
logger = logging.getLogger(__file__)


@app.callback()
def main():
    """
    Manage data.
    """
    pass


@app.command()
def make_registry(
    input_directory: Annotated[
        str,
        typer.Option(
            "--input-directory",
            "-i",
            help="Path to input directory.",
        ),
    ] = ".",
    output_file: Annotated[
        Optional[str],
        typer.Option(
            "--output-file",
            "-o",
            help="Path to output file (default: '<input_directory>/registry.txt').",
        ),
    ] = None,
    rules: Annotated[
        Optional[str],
        typer.Option(
            "--rules",
            "-r",
            help="Path to the registry rule file "
            "(default: '<input_directory>/registry_rules.yml').",
        ),
    ] = None,
    hash_algorithm: Annotated[
        str,
        typer.Option(
            "--hash-algorithm",
            "-a",
            help="Hashing algorithm (default: 'sha256').",
        ),
    ] = "sha256",
):
    """
    Recursively construct a file registry from the current working directory.
    """
    from eradiate.data._core import load_rules, make_registry

    input_directory = Path(input_directory)
    console.print(f"Creating registry file from '{input_directory}'")

    # Load include and exclude rules
    if rules is None:
        rules = input_directory / "registry_rules.yml"
    console.print(f"Using rules in '{rules}'")
    rule_map = load_rules(rules)

    # Write registry
    if output_file is None:
        output_file = input_directory / "registry.txt"
    console.print(f"Writing registry file to '{output_file}'")
    make_registry(
        input_directory,
        output_file,
        includes=rule_map["include"],
        excludes=rule_map["exclude"],
        alg=hash_algorithm,
        show_progress=True,
    )


@app.command()
def update_registries():
    """
    Update local registries for online sources.
    """
    # Update data store registries
    for data_store_id, data_store in eradiate.data.data_store.stores.items():
        if isinstance(data_store, eradiate.data.SafeOnlineDataStore):
            console.print(
                f"[bold cyan]{data_store_id}[/] [{data_store.__class__.__name__}]"
            )
            console.print(f"    Refreshing registry \[{data_store.registry_path}]")
            data_store.registry_delete()
            console.print(
                "    Downloading from "
                f"\[{os.path.join(data_store.base_url, data_store.registry_fname)}]"
            )
            data_store.registry_fetch()
            console.print("  [green]✓ Done[/]")
            continue


@app.command()
def fetch(
    file_list: Annotated[
        Optional[List[str]],
        typer.Argument(
            help="An arbitrary number of relative paths to files to be "
            "retrieved from the data store or file list specifications. "
            "A file list keyword can also be specified to download a "
            "pre-defined list of files (use the `--list` option to display) "
            "available keywords. "
            "If unset, the list of files is read from a YAML file which can be "
            "specified by using the `--from-file` option, if it is used. "
            "Otherwise, it defaults to `all` in a production environment and "
            "`minimal` in a development environment."
        ),
    ] = None,
    from_file: Annotated[
        Optional[str],
        typer.Option(
            "--from-file",
            "-f",
            help="Optional path to a file list (YAML format). If this option "
            "is set, the FILES argument(s) will be ignored.",
        ),
    ] = None,
    list: Annotated[
        bool,
        typer.Option(
            "--list",
            "-l",
            help="List built-in file lists and exit.",
        ),
    ] = False,
):
    """
    Fetch files from the Eradiate data store.
    """
    if list:
        print("Known file lists:")
        for file_list in known_file_lists():
            print(f"  {file_list}")
        exit(0)

    if not file_list:
        if from_file is None:
            if eradiate.config.SOURCE_DIR is None:
                file_list = ["all"]
            else:
                file_list = ["minimal"]
        else:
            console.print(f"Reading file list from '{from_file}'")
            yaml = YAML()
            file_list = yaml.load(Path(from_file))

    converted = []
    for spec in file_list:
        try:
            converted.extend(get_file_list(spec))
        except ValueError:
            converted.append(spec)

    file_list = converted

    for filename in file_list:
        try:
            console.print(f"[blue]Fetching '{filename}'[/]")
            path = eradiate.data.data_store.fetch(filename)
        except DataError:
            console.print("[red]✗[/] not found")
        else:
            console.print(f"[green]✓[/] found \[{path}]")


@app.command()
def purge_cache(
    keep: Annotated[
        bool, typer.Option("--keep", "-k", help="Keep registered files.")
    ] = False,
):
    """
    Purge the cache of online data stores.
    """
    for data_store_id, data_store in eradiate.data.data_store.stores.items():
        console.print(
            f"[bold cyan]{data_store_id}[/] [{data_store.__class__.__name__}]"
        )

        if isinstance(data_store, eradiate.data.SafeDirectoryDataStore):
            console.print("  Skipping")
            continue

        if isinstance(data_store, eradiate.data.SafeOnlineDataStore):
            console.print(f"  Purging '{data_store.path}'")
            if keep:
                data_store.purge(keep="registered")
            else:
                data_store.purge()
            continue

        if isinstance(data_store, eradiate.data.BlindOnlineDataStore):
            console.print(f"  Purging '{data_store.path}'")
            data_store.purge()
            continue


@app.command()
def info(
    data_stores: Annotated[
        Optional[List[str]],
        typer.Argument(
            help=(
                "List of data stores for which information is requested. "
                "If no data store ID is passed, information is displayed for "
                "all data stores."
            )
        ),
    ] = None,
    list_registry: Annotated[
        bool,
        typer.Option(
            "-l", "--list-registry", help="Show registry content if relevant."
        ),
    ] = False,
):
    """
    Display information about data store configuration.
    """

    # Build section list
    sections = [
        ("[purple]Base URL[/] \[base_url]", "base_url"),
        ("[purple]Local path[/] \[path]", "path"),
        ("[purple]Registry path[/] \[registry_path]", "registry_path"),
    ]

    if list_registry:
        sections.append(("Registered files", "registry_keys"))

    first = True

    for data_store_id, data_store in eradiate.data.data_store.stores.items():
        if data_stores and (data_store_id not in data_stores):
            continue

        # Collect section contents
        reprs = {}

        for attr in ["base_url", "path", "registry_path"]:
            try:
                reprs[attr] = f"'{getattr(data_store, attr)}'"
            except (NotImplementedError, AttributeError):
                reprs[attr] = None

        try:
            reprs["registry_keys"] = "\n".join(
                f"'{x}'" for x in sorted(data_store.registry.keys())
            )
        except (NotImplementedError, AttributeError):
            reprs["registry_keys"] = None

        # Display the content for current data store
        if not first:
            console.print()
        else:
            first = False

        console.print(
            f"[bold cyan]{data_store_id}[/] \[{data_store.__class__.__name__}]"
        )

        for section_title, repr_key in sections:
            the_repr = reprs[repr_key]

            if the_repr is not None:
                console.print(f"  {section_title}")
                console.print(textwrap.indent(reprs[repr_key], " " * 4))


@app.command()
def check(
    keywords: Annotated[
        Optional[List[str]],
        typer.Argument(
            help="A keyword defining the datasets that are to be checked. "
            "See the `--list` option for available keywords,"
        ),
    ] = None,
    list: Annotated[
        bool,
        typer.Option(
            "--list",
            "-l",
            help="List known keywords and exit.",
        ),
    ] = False,
    fix: Annotated[
        bool,
        typer.Option(
            "--fix",
            "-f",
            help="Fix issues that can be.",
        ),
    ] = False,
):
    """
    Check data for availability and integrity, optionally fix them.
    """
    if list:
        print("Known keywords:")
        for key in KNOWN_MOLECULAR_ABSORPTION_DATABASES.keys():
            print(f"  {key}")
        exit(0)

    for key in keywords:
        logger.info(f"Opening '{key}'")
        try:
            AbsorptionDatabase.from_name(key, fix=fix)
            logger.info("Success!")
        except FileNotFoundError:
            pass



import attrs

from ._core import PhaseFunction
from ...attrs import parse_docs


@parse_docs
@attrs.define(eq=False, slots=False)
class RayleighPhaseFunction(PhaseFunction):
    """
    Rayleigh phase function [``rayleigh``].

    The Rayleigh phase function models scattering by particles with a
    characteristic size much smaller than the considered radiation wavelength.
    It is typically used to represent scattering by gas molecules in the
    atmosphere.
    """

    @property
    def template(self) -> dict:
        return {"type": "rayleigh"}



from __future__ import annotations

import enum
import functools
import typing as t

import attrs
import mitsuba

from .attrs import documented, parse_docs
from .exceptions import UnsupportedModeError

# ------------------------------------------------------------------------------
#                                 Mode flags
# ------------------------------------------------------------------------------


class Flag(enum.Flag):
    """
    Small extension to :class:`enum.Flag` that adds a ``convert()`` class method
    constructor.
    """

    @classmethod
    def convert(cls, value: t.Any) -> Flag:
        """
        Try to convert a value to a flag. Strings are capitalized and converted
        to the corresponding enum member.
        """
        if isinstance(value, str):
            return cls[value.upper()]
        else:
            return cls(value)


class SpectralMode(Flag):
    """
    Spectral dimension handling flags.
    """

    MONO = enum.auto()  #: Monochromatic (line-by-line) mode
    CKD = enum.auto()  #: Correlated-k distribution mode


_SPECTRAL_COORD_LABELS = {SpectralMode.MONO: "w", SpectralMode.CKD: "bd"}


class MitsubaBackend(Flag):
    """
    Mitsuba backend flags.
    """

    SCALAR = enum.auto()  #: Scalar backend
    LLVM = enum.auto()  #: LLVM backend


class MitsubaColorMode(Flag):
    """
    Mitsuba color mode flags.
    """

    MONO = enum.auto()  #: Monochromatic mode


# ------------------------------------------------------------------------------
#                              Mode definitions
# ------------------------------------------------------------------------------


# Map associating a mode ID string to the corresponding class
# (aliased in public API section). This is implemented as a cached function to
# make a delayed evaluation possible (otherwise, the Mode class is not defined).
# See also the Mode.new() constructor.
@functools.lru_cache(maxsize=1)
def _mode_registry() -> dict[str, Mode]:
    return {
        k: Mode(id=k, **v)
        for k, v in {
            "mono_single": {
                "spectral_mode": SpectralMode.MONO,
                "mi_backend": MitsubaBackend.SCALAR,
                "mi_color_mode": MitsubaColorMode.MONO,
                "mi_double_precision": False,
                "mi_polarized": False,
            },
            "mono_double": {
                "spectral_mode": SpectralMode.MONO,
                "mi_backend": MitsubaBackend.SCALAR,
                "mi_color_mode": MitsubaColorMode.MONO,
                "mi_double_precision": True,
                "mi_polarized": False,
            },
            "ckd_single": {
                "spectral_mode": SpectralMode.CKD,
                "mi_backend": MitsubaBackend.SCALAR,
                "mi_color_mode": MitsubaColorMode.MONO,
                "mi_double_precision": False,
                "mi_polarized": False,
            },
            "ckd_double": {
                "spectral_mode": SpectralMode.CKD,
                "mi_backend": MitsubaBackend.SCALAR,
                "mi_color_mode": MitsubaColorMode.MONO,
                "mi_double_precision": True,
                "mi_polarized": False,
            },
        }.items()
    }


@parse_docs
@attrs.frozen
class Mode:
    """
    Data structure describing Eradiate's operational mode and associated
    ancillary data.

    Warnings
    --------
    Instances are immutable.
    """

    id: str = documented(
        attrs.field(converter=attrs.converters.optional(str)),
        doc="Mode identifier.",
        type="str",
    )

    spectral_mode: SpectralMode = documented(
        attrs.field(converter=SpectralMode.convert),
        doc="Spectral dimension handling.",
        type=":class:`.SpectralMode`",
        init_type=":class:`.SpectralMode` or str",
    )

    mi_backend: MitsubaBackend = documented(
        attrs.field(converter=MitsubaBackend.convert),
        doc="Mitsuba computational backend.",
        type=":class:`.MitsubaBackend`",
        init_type=":class:`.MitsubaBackend` or str",
    )

    mi_color_mode: MitsubaColorMode = documented(
        attrs.field(converter=MitsubaColorMode.convert),
        doc="Mitsuba color mode.",
        type=".MitsubaColorMode",
        init_type=":class:`.MitsubaColorMode` or str",
    )

    mi_polarized: bool = documented(
        attrs.field(default=False, converter=bool),
        doc="Mitsuba polarized mode.",
        type="bool",
        default="False",
    )

    mi_double_precision: bool = documented(
        attrs.field(default=True, converter=bool),
        doc="Mitsuba double precision.",
        type="bool",
        default="True",
    )

    @property
    def spectral_coord_label(self) -> str:
        """
        Spectral coordinate label.
        """
        return _SPECTRAL_COORD_LABELS[self.spectral_mode]

    @property
    def mi_variant(self):
        """
        Mitsuba variant associated with the selected mode.
        """

        result = [self.mi_backend.name.lower(), self.mi_color_mode.name.lower()]
        if self.mi_polarized:
            result.append("polarized")
        if self.mi_double_precision:
            result.append("double")
        return "_".join(result)

    def check(
        self,
        spectral_mode: None | SpectralMode | str = None,
        mi_backend: None | MitsubaBackend | str = None,
        mi_color_mode: None | MitsubaColorMode | str = None,
        mi_polarized: bool | None = None,
        mi_double_precision: bool | None = None,
    ) -> bool:
        """
        Check if the currently active mode has the passed flags.

        Parameters
        ----------
        spectral_mode : :class:`.SpectralMode` or str, optional
            Spectral mode to check. If unset, the check is skipped.

        mi_backend : :class:`.MitsubaBackend` or str, optional
            Mitsuba backend to check. If unset, the check is skipped.

        mi_color_mode : :class:`.MitsubaColorMode` or str, optional
            Mitsuba color mode to check. If unset, the check is skipped.

        mi_polarized : bool, optional
            Mitsuba polarized mode to check. If unset, the check is skipped.

        mi_double_precision : bool, optional
            Mitsuba double precision mode to check. If unset, the check is skipped.

        Returns
        -------
        bool
            ``True`` if current mode has the passed flags, ``False`` otherwise.
        """
        outcome = True

        if spectral_mode is not None:
            outcome &= bool(self.spectral_mode & SpectralMode.convert(spectral_mode))

        if mi_backend is not None:
            outcome &= bool(self.mi_backend & MitsubaBackend.convert(mi_backend))

        if mi_color_mode is not None:
            outcome &= bool(
                self.mi_color_mode & MitsubaColorMode.convert(mi_color_mode)
            )

        if mi_polarized is not None:
            outcome &= self.mi_polarized == mi_polarized

        if mi_double_precision is not None:
            outcome &= self.mi_double_precision == mi_double_precision

        return outcome

    @property
    def is_mono(self) -> bool:
        return self.spectral_mode is SpectralMode.MONO

    @property
    def is_ckd(self) -> bool:
        return self.spectral_mode is SpectralMode.CKD

    @property
    def is_single_precision(self) -> bool:
        return self.mi_double_precision is False

    @property
    def is_double_precision(self) -> bool:
        return self.mi_double_precision is True

    @staticmethod
    def new(mode_id: str) -> Mode:
        """
        Create a :class:`Mode` instance given its identifier. Available modes are:

        * ``mono_single``: Monochromatic, single-precision
        * ``mono_double``: Monochromatic, double-precision
        * ``mono``: Alias to ``mono_double``
        * ``ckd_single``: CKD, single-precision
        * ``ckd_double``: CKD, double-precision
        * ``ckd``: Alias to ``ckd_double``

        Parameters
        ----------
        mode_id : str
            String identifier for the created :class:`Mode` instance.

        Returns
        -------
        .Mode
            Created :class:`.Mode` instance.
        """
        try:
            return _mode_registry()[mode_id]
        except KeyError:
            raise ValueError(f"unknown mode '{mode_id}'")


# Define mode aliases
_mode_registry().update(
    {
        "mono": _mode_registry()["mono_double"],
        "ckd": _mode_registry()["ckd_double"],
    }
)

# Eradiate's operational mode configuration
_active_mode: Mode | None = None


# ------------------------------------------------------------------------------
#                                Public API
# ------------------------------------------------------------------------------


def mode() -> Mode | None:
    """
    Get current operational mode.

    Returns
    -------
    .Mode or None
        Current operational mode.
    """
    return _active_mode


def modes(filter: t.Callable[[Mode], bool] | None = None) -> dict[str, Mode]:
    """
    Get list of registered operational modes.

    Parameters
    ----------
    filter : callable, optional
        A callable used to filter the returned modes. Operates on a
        :class:`.Mode` instance.

    Returns
    -------
    modes: dict[str, .Mode]
        List of registered operational modes.

    Examples
    --------
    Return the full list of registered modes:

    >>> eradiate.modes()

    Return only CKD modes:

    >>> eradiate.modes(lambda x: x.is_ckd)
    """
    if filter is None:
        filter = lambda x: True

    return {k: v for k, v in _mode_registry().items() if filter(v)}


def set_mode(mode_id: str):
    """
    Set Eradiate's operational mode.

    This function sets and configures Eradiate's operational mode. Eradiate's
    modes map to Mitsuba's variants and are used to make contextual decisions
    when relevant during the translation of a scene to its kernel format.

    .. admonition:: Valid mode IDs
       :class: info

       * ``mono`` (monochromatic mode, single precision)
       * ``mono_double`` (monochromatic mode, double-precision)
       * ``ckd`` (CKD mode, single precision)
       * ``ckd_double`` (CKD mode, double-precision)
       * ``none`` (no mode selected)

    Parameters
    ----------
    mode_id : str
        Mode to be selected (see list below).

    Raises
    ------
    ValueError
        ``mode_id`` does not match any of the known mode identifiers.
    """
    global _active_mode

    if mode_id in _mode_registry():
        mode = Mode.new(mode_id)
        mitsuba.set_variant(mode.mi_variant)
    elif mode_id.lower() == "none":
        mode = None
    else:
        raise ValueError(f"unknown mode '{mode_id}'")

    _active_mode = mode


def supported_mode(**kwargs):
    """
    Check whether the current mode has specific features. If not, raise.

    Parameters
    ----------
    kwargs
        Keyword arguments passed to :meth:`.Mode.check`.

    Raises
    ------
    .UnsupportedModeError
        Current mode does not pass the check.
    """
    if mode() is None or not mode().check(**kwargs):
        raise UnsupportedModeError


def unsupported_mode(**kwargs):
    """
    Check whether the current mode has specific features. If so, raise.

    Parameters
    ----------
    kwargs
        Keyword arguments passed to :meth:`.Mode.check`.

    Raises
    ------
    .UnsupportedModeError
        Current mode has the requested flags.
    """
    if mode() is None or mode().check(**kwargs):
        raise UnsupportedModeError



"""
Post-processing pipeline access points.
"""

from __future__ import annotations

import importlib

from hamilton import telemetry
from hamilton.base import DictResult, SimplePythonGraphAdapter
from hamilton.driver import Driver
from rich.table import Table

import eradiate

from .._mode import Mode
from ..scenes.integrators import Integrator
from ..scenes.measure import Measure
from ..scenes.spectra import InterpolatedSpectrum

# Disable Hamilton telemetry (we don't want to bother our users with it)
telemetry.disable_telemetry()


def config(
    measure: Measure,
    mode: Mode | str | None = None,
    integrator: Integrator | None = None,
) -> dict:
    """
    Generate a pipeline configuration for a specific scene setup.

    Parameters
    ----------
    measure : .Measure

    mode : .Mode or str, optional
        Mode or mode ID for which the pipeline is configured. By default, the
        current active mode is used.

    integrator : .Integrator or None, optional
        Integrator used for the experiment; indicates whether the moment was
        calculated during the integration.

    Returns
    -------
    dict

    See Also
    --------
    :mod:`eradiate.pipelines`
    """
    result = {}

    # Which mode is selected?
    if mode is None:
        mode = eradiate.mode()
    if isinstance(mode, str):
        mode = eradiate.Mode.new(mode)
    result["mode_id"] = mode.id

    # Is the measure distant?
    result["measure_distant"] = measure.is_distant()

    # Does the measure provide viewing angle values?
    result["add_viewing_angles"] = hasattr(measure, "viewing_angles")

    # Which physical variable are we processing?
    result["var_name"], result["var_metadata"] = measure.var

    # Shall we apply spectral response function weighting (a.k.a convolution)?
    result["apply_spectral_response"] = isinstance(measure.srf, InterpolatedSpectrum)

    # Should we calculate the variance in the result?
    result["calculate_variance"] = integrator.moment if integrator is not None else False

    return result


def driver(config: dict):
    """
    Create a Hamilton :class:`~hamilton.driver.Driver` instance using the
    specified configuration. The post-processing pipeline mutates defined on
    parameters.

    Parameters
    ----------
    config : dict
        A configuration dictionary specifying various parameters
        (see :mod:`~eradiate.pipelines` for details).

    Returns
    -------
    hamilton.Driver

    See Also
    --------
    :func:`.config`
    """
    # Force power user mode, required to mutate node names based on the
    # processed variable name
    config = {"hamilton.enable_power_user_mode": True, **config}

    # Instantiate the driver following the Generic result builder documentation
    # https://hamilton.dagworks.io/en/latest/reference/result-builders/Generic/
    module = importlib.import_module("eradiate.pipelines.definitions")
    dict_builder = DictResult()
    adapter = SimplePythonGraphAdapter(dict_builder)
    drv = Driver(config, module, adapter=adapter)
    return drv


def list_variables(
    drv: Driver, as_table: bool = False
) -> list[tuple[str, str, str]] | Table:
    """
    List variables available in a pipeline. Configuration variables are omitted.

    Parameters
    ----------
    drv : hamilton.driver.Driver
         Pipeline driver to inspect.

    as_table : bool, default: False
        If ``True``, format the results as a table.

    Returns
    -------
    variables : list or rich.table.Table
        List of variables ordered by name. Each element contains the following
        entries:

        * name;
        * type;
        * whether the variable is an input parameter.
    """
    data = []

    for variable in drv.list_available_variables():
        if variable.name in drv.graph.config:
            continue

        line = (
            variable.name,
            variable.type.__name__,
            "x" if variable.is_external_input else "",
        )
        data.append(line)

    result = sorted(data, key=lambda x: (0 if x[2] == "x" else 1, x[0]))

    if not as_table:
        return result
    else:
        table = Table()
        table.add_column("Name")
        table.add_column("Type")
        table.add_column("Input")

        for line in result:
            table.add_row(*line)

        return table


def outputs(drv: Driver):
    """
    Generate a list of default outputs for a given pipeline. Only nodes
    with the ``"final": "true"`` and ``"kind": "data"`` tags are selected.
    """
    return [
        x.name
        for x in drv.list_available_variables()
        if x.tags.get("final") == "true" and x.tags.get("kind") == "data"
    ]



from __future__ import annotations

__all__ = [
    "auto_or",
    "convert_thermoprops",
    "on_quantity",
    "to_dataset",
]

import os
import typing as t

import mitsuba as mi
import numpy as np
import pint
import xarray as xr

from . import data
from .attrs import AUTO
from .data import data_store
from .typing import PathLike


def on_quantity(
    wrapped_converter: t.Callable[[t.Any], t.Any],
) -> t.Callable[[t.Any], t.Any]:
    """
    Apply a converter to the magnitude of a :class:`pint.Quantity`.

    Parameters
    ----------
    wrapped_converter : callable
        The converter which will be applied to the magnitude of a
        :class:`pint.Quantity`.

    Returns
    -------
    callable
    """

    def f(value: t.Any) -> t.Any:
        if isinstance(value, pint.Quantity):
            return wrapped_converter(value.magnitude) * value.units
        else:
            return wrapped_converter(value)

    return f


def auto_or(
    wrapped_converter: t.Callable[[t.Any], t.Any],
) -> t.Callable[[t.Any], t.Any]:
    """
    A converter that allows an attribute to be set to :data:`.AUTO`.

    Parameters
    ----------
    wrapped_converter : callable
        The converter that is used for non-:data:`.AUTO` values.

    Returns
    -------
    callable
    """

    def f(value):
        if value is AUTO:
            return value

        return wrapped_converter(value)

    return f


def to_dataset(
    load_from_id: t.Callable[[str], xr.Dataset] | None = None,
) -> t.Callable[[xr.Dataset | PathLike], xr.Dataset]:
    """
    Generates a converter that converts a value to a :class:`xarray.Dataset`.

    Parameters
    ----------
    load_from_id : callable, optional
        A callable with the signature ``f(x: str) -> Dataset`` used to
        interpret dataset identifiers.
        Set this parameter to handle dataset identifiers.
        If unset, dataset identifiers are not supported.

    Returns
    -------
    A dataset converter.

    Notes
    -----
    The conversion logic is as follows:

    1. If the value is an xarray dataset, it is returned directly.
    2. If the value is a path-like object ending with the ``.nc`` extension, the
       converter tries to load a dataset from that location, first locally, then
       (should that fail) from the Eradiate data store.
    3. If the value is a string and ``load_from_id`` is not ``None``, it is
        interpreted as a dataset identifier and ``load_from_id(value)`` is
        returned.
    4. Otherwise, a :class:`ValueError` is raised.

    Examples
    --------
    A converter with basic dataset identifier interpretation (the passed
    callable may implement more complex logic, *e.g.* with identifier
    fallback substitution):

    >>> aerosol_converter = to_dataset(
    ...     lambda x: data.load_dataset(f"spectra/particles/{x}.nc")
    ... )

    A converter without dataset identifier interpretation:

    >>> aerosol_converter = to_dataset()
    """

    def converter(value: xr.Dataset | PathLike) -> xr.Dataset:
        if isinstance(value, xr.Dataset):
            return value

        # Path (local or remote)
        if str(value).endswith(".nc"):
            # Try and open a file if it is directly referenced
            if os.path.isfile(value):
                return xr.load_dataset(value)

            # Try and serve the file from the data store
            return data.load_dataset(value)

        # Identifier for a dataset in the data store
        if isinstance(value, str) and load_from_id is not None:
            return load_from_id(value)

        # Abnormal state
        # Reference must be provided as a Dataset, a path-like or a str
        raise ValueError(f"Cannot convert value '{value}'")

    return converter


def to_mi_scalar_transform(value):
    """
    Convert an array-like value to a :class:`mitsuba.ScalarTransform4f`.
    If `value` is a Numpy array, it is used to initialize a
    :class:`mitsuba.ScalarTransform4f` without copy; if it is a list, a Numpy
    array is first created from it. Otherwise, `value` is forwarded without
    change.
    """
    if isinstance(value, np.ndarray):
        return mi.ScalarTransform4f(value)

    elif isinstance(value, list):
        return mi.ScalarTransform4f(np.array(value))

    else:
        return value


def convert_thermoprops(value) -> xr.Dataset:
    """Converter for atmosphere thermophysical properties specifications."""
    import joseki

    # Dataset: do nothing
    if isinstance(value, xr.Dataset):
        return value

    # PathLike: try to load dataset
    elif isinstance(value, (os.PathLike, str)):
        path = data_store.fetch(value)
        if path.is_file():
            return joseki.load_dataset(path)
        else:
            raise ValueError(
                f"invalid path for 'thermoprops': {path} (expected a file)"
            )

    # Dictionary: forward to joseki.make()
    elif isinstance(value, dict):
        return joseki.make(**value)

    # Anything else: raise error
    else:
        raise TypeError(
            f"invalid type for 'thermoprops': {type(value)} "
            f"(expected Dataset or PathLike)"
        )


def _is_quantity(x):
    return isinstance(x, pint.Quantity)


def _isinstance_of_2tuple_quantity(x):
    return isinstance(x, tuple) and list(map(_is_quantity, x)) == [True, True]



from ._atmosphere import AtmosphereExperiment as AtmosphereExperiment
from ._canopy import CanopyExperiment as CanopyExperiment
from ._canopy_atmosphere import CanopyAtmosphereExperiment as CanopyAtmosphereExperiment
from ._core import EarthObservationExperiment as EarthObservationExperiment
from ._core import Experiment as Experiment
from ._core import run as run
from ._dem import DEMExperiment as DEMExperiment



"""
Extensions and helpers for tutorials.
"""

from __future__ import annotations

import matplotlib.pyplot as plt
import xarray as xr

from ..scenes.atmosphere import AbstractHeterogeneousAtmosphere
from ..spectral.index import SpectralIndex


def plot_sigma_t(
    *atmospheres: AbstractHeterogeneousAtmosphere,
    labels: list[str] | None = None,
    altitude_extent: tuple[float, float] | None = None,
    si: SpectralIndex | None = None,
    show: bool = True,
) -> tuple[plt.Figure, plt.Axes]:
    """
    Display the extinction coefficient of one or several atmosphere objects for
    a single spectral context.

    Parameters
    ----------
    *atmospheres : .AbstractHeterogeneousAtmosphere
        One or several atmosphere objects for which to plot the extinction
        coefficient.

    labels : list of strings, optional
        Labels associated with the passed atmosphere objects. If unset, no
        legend will be added to the plot.

    altitude_extent : tuple of float
        A (min, max) altitude pair (in km) to which the plot is restricted.

    si : .SpectralIndex, optional
        The spectral index at which the extinction coefficient is evaluated.
        If unset, a default spectral index is created using
        :meth:`.SpectralIndex.new`.

    show : bool, optional
        If ``True``, return ``None`` and display the plot. Otherwise, return a
        (``Figure``, ``Axes``) pair.
    """

    from matplotlib.ticker import ScalarFormatter

    from eradiate.units import to_quantity

    if si is None:  # If none is found, fall back to the default (550 nm)
        si = SpectralIndex.new()
    else:  # Otherwise, apply conversion protocol
        si = SpectralIndex.convert(si)

    if labels is None:
        label_iter = iter([None for _ in atmospheres])
    else:
        label_iter = iter(labels)

    fig, ax = plt.subplots(1, 1)

    with plt.rc_context({"lines.linestyle": ":", "lines.marker": "."}):
        for atmosphere in atmospheres:
            altitude = to_quantity(atmosphere.eval_radprops(si=si).z_layer).m_as("km")
            sigma_t = to_quantity(atmosphere.eval_radprops(si=si).sigma_t).m_as("1/m")
            ax.plot(altitude, sigma_t, label=next(label_iter))

    formatter = ScalarFormatter(useMathText=True)
    formatter.set_powerlimits((-3, 2))
    ax.yaxis.set_major_formatter(formatter)
    ax.set_ylabel("Extinction coefficient [1/m]")

    if labels is not None:
        fig.legend(
            bbox_to_anchor=(1.0, 0.5),
            loc="center left",
            borderaxespad=0.0,
        )

    if altitude_extent is not None:
        ax.set_xlim(altitude_extent)
    ax.set_xlabel("Altitude [km]")

    plt.tight_layout()
    if show:
        plt.show()
    else:
        return fig, ax


def load_ipython_extension(ipython):
    """
    IPython extension for Eradiate tutorials. Should be loaded at the top of the
    tutorial notebook.

    It sets the Matplotlib style and prints the current date and Eradiate
    version as markdown.

    See Also
    --------
    :func:`eradiate.plot.set_style`

    Notes
    -----

    This extension should be loaded using the IPython magic:

    .. code::

       %load_ext eradiate.notebook.tutorials
    """

    import datetime

    from IPython.display import Markdown, display

    from .. import __version__
    from ..plot import set_style

    set_style()
    xr.set_options(display_expand_data=False)
    display(
        Markdown(
            f"*Last updated: {datetime.datetime.now():%Y-%m-%d %H:%M} "
            f"(eradiate v{__version__})*"
        )
    )



from __future__ import annotations

import os
from pathlib import Path


def _validate_source_dir(value: Path | None):
    if SOURCE_DIR is None:
        # Import must be local and not use the lazy loader to avoid circular imports
        from ..kernel._versions import kernel_installed

        kernel_is_installed, _ = kernel_installed()

        if not kernel_is_installed:
            raise RuntimeError(
                "Could not find a suitable production installation for the "
                "Eradiate kernel. This is either because you are using Eradiate "
                "in a production environment without having the eradiate-mitsuba "
                "package installed, or because you are using Eradiate directly "
                "from the sources. In the latter case, please make sure the "
                "'ERADIATE_SOURCE_DIR' environment variable is correctly set to "
                "the Eradiate installation directory. If you are using Eradiate "
                "directly from the sources, you can alternatively source the "
                "provided setpath.sh script. You can install the eradiate-mitsuba "
                "package using 'pip install eradiate-mitsuba'."
            )

    else:
        eradiate_init = value / "src" / "eradiate" / "__init__.py"

        if not eradiate_init.is_file():
            raise RuntimeError(
                f"While configuring Eradiate: could not find {eradiate_init} file. "
                "Please make sure the 'ERADIATE_SOURCE_DIR' environment variable is "
                "correctly set to the Eradiate installation directory. If you are "
                "using Eradiate directly from the sources, you can alternatively "
                "source the provided setpath.sh script. If you wish to use Eradiate "
                "in a production environment, you can install the eradiate-mitsuba "
                "package using 'pip install eradiate-mitsuba' and unset the "
                "'ERADIATE_SOURCE_DIR' environment variable."
            ) from FileNotFoundError(eradiate_init)


#: Path to the Eradiate source code directory, if relevant. Takes the value of
#: the ``ERADIATE_SOURCE_DIR`` environment variable if it is set; otherwise
#: defaults to ``None``.
SOURCE_DIR: Path | None = os.environ.get("ERADIATE_SOURCE_DIR", default=None)
if SOURCE_DIR is not None:
    SOURCE_DIR = Path(SOURCE_DIR).resolve()
_validate_source_dir(SOURCE_DIR)

#: Identifier of the environment in which Eradiate is used. Takes the value of
#: the ``ERADIATE_ENV`` environment variable if it is set; otherwise defaults to
#: ``"default"``.
ENV: str = os.environ.get("ERADIATE_ENV", "default")



from __future__ import annotations

import attrs
import numpy as np
import pint
import pinttr

from ._core import Spectrum
from ... import validators
from ...attrs import documented, parse_docs
from ...spectral.ckd import BinSet
from ...spectral.mono import WavelengthSet
from ...units import unit_context_kernel as uck
from ...util.misc import summary_repr_vector


@parse_docs
@attrs.define(eq=False, slots=False)
class MultiDeltaSpectrum(Spectrum):
    """
    A spectrum made of multiple translated Dirac delta [``multi_delta``].

    Notes
    -----
    This spectrum is intended to be used for spectral grid specification.
    It has no kernel-level representation, and it cannot be evaluated.
    """

    wavelengths: pint.Quantity = documented(
        pinttr.field(
            kw_only=True,
            units=uck.deferred("wavelength"),
            converter=[
                lambda x: np.sort(np.atleast_1d(x)),
                pinttr.converters.to_units(uck.deferred("wavelength")),
                lambda x: np.unique(x.m) * x.units,
            ],
            validator=validators.all_strictly_positive,
            repr=lambda x: f"{summary_repr_vector(x.m)} {x.u:~}",
        ),
        doc="An array of wavelengths specifying the translation wavelength of each "
        "Dirac delta. Wavelength values are positive and unique. "
        "When a single value is provided, it is converted to a 1-element array. "
        "Wavelength are deduplicated and sorted by ascending values. "
        'Unit-enabled field (default: ``ucc["wavelength"]``).',
        type="quantity",
        init_type="array-like or quantity",
    )

    def eval_mono(self, w: pint.Quantity) -> pint.Quantity:
        raise NotImplementedError

    def eval_ckd(self, w: pint.Quantity, g: float) -> pint.Quantity:
        raise NotImplementedError

    def integral(self, wmin: pint.Quantity, wmax: pint.Quantity) -> pint.Quantity:
        raise NotImplementedError

    @property
    def template(self) -> dict:
        raise NotImplementedError

    @property
    def params(self) -> dict:
        raise NotImplementedError

    def select_in_wavelength_set(self, wset: WavelengthSet) -> WavelengthSet:
        # the input wavelength set is completely ignored
        # only the attribute wavelengths are included in the returned
        # wavelength set

        return WavelengthSet(self.wavelengths)

    def select_in_bin_set(self, binset: BinSet) -> BinSet:
        bins = binset.bins
        wunits = "nm"
        xmin = np.array([bin.wmin.m_as(wunits) for bin in bins])
        xmax = np.array([bin.wmax.m_as(wunits) for bin in bins])
        x = self.wavelengths.m_as(wunits)
        selected = _select(xmin, xmax, x)
        return BinSet(bins=list(np.array(bins)[selected]))


def _select(xmin, xmax, x):
    selmin = np.searchsorted(xmin, x)
    selmax = np.searchsorted(xmax, x) + 1
    hit = selmin == selmax  # Mask where x values which triggered a bin hit

    # Map x values to selected bin (index -999 means not selected)
    bin_index = np.where(hit, selmin - 1, np.full_like(x, -999)).astype(np.int64)

    # Get selected bins only
    selected = np.unique(bin_index)  # mask removes -999 value
    selected = selected[selected >= 0]

    return selected



from __future__ import annotations

from importlib.resources import files
from pathlib import Path
from typing import Any, Callable

import attrs
from ruamel.yaml import YAML


@attrs.define
class FileListLoader:
    """
    Lazy loader for file lists.
    """

    file_list: list[str] | Callable[[], bool]
    filter: Callable[[Any], bool] = attrs.field(default=lambda x: True)

    def get_file_list(self) -> list[str]:
        file_list = self.file_list() if callable(self.file_list) else self.file_list
        return [path for path in file_list if self.filter(path)]


_yaml = YAML()

FILE_LIST_LOADERS: dict[str, FileListLoader] = {
    "all": FileListLoader(
        lambda: _yaml.load(Path(files("eradiate") / "data/downloads_all.yml"))
    ),
    "minimal": FileListLoader(
        _yaml.load(Path(files("eradiate") / "data/downloads_minimal.yml"))
    ),
}

for name, rule in {
    "komodo": lambda x: x.startswith("spectra/absorption/mono/komodo"),
    "gecko": lambda x: x.startswith("spectra/absorption/mono/gecko"),
    "monotropa": lambda x: x.startswith("spectra/absorption/ckd/monotropa"),
    "mycena": lambda x: x.startswith("spectra/absorption/ckd/mycena"),
    "panellus": lambda x: x.startswith("spectra/absorption/ckd/panellus"),
}.items():
    FILE_LIST_LOADERS[name] = FileListLoader(
        FILE_LIST_LOADERS["all"].file_list, filter=rule
    )


def known_file_lists() -> list[str]:
    return list(FILE_LIST_LOADERS.keys())


def get_file_list(spec: str):
    """
    Get a list of files corresponding to a specifier (i.e. an identifier).
    """
    try:
        return FILE_LIST_LOADERS[spec].get_file_list()
    except KeyError as e:
        raise ValueError(f"unknown file list specification '{spec}'") from e



from __future__ import annotations

import json
import typing as t
from pathlib import Path

import numpy as np

SCENARIO_FILE_NAME = "scenario.json"


def _update_material(
    elem: dict,
    canopy_name: str,
    spectral_data: dict[str, t.Any | dict[str, t.Any]] | None,
) -> dict:
    """
    Update the material of an element with spectral data if available.

    Parameters
    ----------
    elem : dict
        The element to update.
    canopy_name : str
        The name of the canopy or ground.
    spectral_data : dict[str, t.Any | dict[str, t.Any]] | None
        The spectral data to apply to the element.

    Returns
    -------
    dict
        The updated element.
    """
    elem_copy = elem.copy()
    if (
        spectral_data is not None
        and canopy_name in spectral_data
        and (canopy_name == "ground" or elem["id"] in spectral_data[canopy_name])
    ):
        assert isinstance(spectral_data, dict)
        try:
            del elem_copy["reflectance"]
        except KeyError:
            pass
        try:
            del elem_copy["transmittance"]
        except KeyError:
            pass
        if canopy_name == "ground":
            return {**elem_copy, **spectral_data[canopy_name]}
        else:
            assert isinstance(spectral_data[canopy_name], dict)
            return {**elem_copy, **spectral_data[canopy_name][elem["id"]]}
    else:
        return elem


def _parse_rpv_surface(
    surface: dict[str, t.Any], spectral_data: dict[str, t.Any]
) -> dict[str, t.Any]:
    """
    Convert surface data to RPV representation.

    Parameters
    ----------
    surface : dict
        Surface data to convert.
    spectral_data : dict
        Spectral data to apply to the surface.

    Returns
    -------
    dict
        RPV representation of the surface data.
    """
    return {
        **surface,
        **_update_material(
            {"rho_0": surface.get("rho_0", {})},
            "ground",
            spectral_data,
        ),
        **_update_material(
            {"g": surface.get("g", {})},
            "ground",
            spectral_data,
        ),
        **_update_material(
            {"k": surface.get("k", {})},
            "ground",
            spectral_data,
        ),
    }


def _parse_lambertian_surface(
    surface: dict[str, t.Any], spectral_data: dict[str, t.Any]
) -> dict[str, t.Any]:
    """
    Convert surface data to reflectance representation.

    Parameters
    ----------
    surface : dict
        Surface data to convert.
    spectral_data : dict
        Spectral data to apply to the surface.

    Returns
    -------
    dict
        Reflectance representation of the surface data.
    """
    return {
        **surface,
        **_update_material(
            {"reflectance": surface.get("reflectance", {})},
            "ground",
            spectral_data,
        ),
        **(
            _update_material(
                {"transmittance": surface.get("transmittance", {})},
                "ground",
                spectral_data,
            )
            if "transmittance" in surface
            else {}
        ),
    }


def apply_transformation(transf: np.ndarray, center: np.ndarray) -> np.ndarray:
    """
    Apply transformation matrix to origin and translate the instance positions to the center of the scenario.

    Parameters
    ----------
    transf : np.ndarray
        Array of transformation values.
    center : np.ndarray
        Array representing the center of the scenario.

    Returns
    -------
    np.ndarray
        Transformed positions with center adjustment.
    """
    origin = np.array([0.0, 0.0, 0.0, 1.0]).T

    # Apply the transformation matrix to origin and adjust the center
    return (transf @ origin)[:3].T - center


def load_scenario(
    scenario_folder: Path,
    padding: int,
    spectral_data: dict[str, t.Any | dict[str, t.Any]] | None = None,
) -> dict:
    """
    Parse JSON file of scenario from a given path. Apply transformation to the
    data by converting units of wavelengths and compute instance positions.

    Parameters
    ----------
    scenario_folder : path-like
        Path of the folder containing scenario JSON file.
    padding : int
        Padding to apply to the scenario.
    spectral_data : dict[str, t.Any or dict[str, t.Any]] or None
        Spectral data to apply to the scenario, defaults to None (keep original).
        Example:

        .. code:: python

            spectral_data = {
                "ground": {
                    "reflectance": ground_reflectance,
                },
                "object_name": {
                    "subobject_name": {
                        # Spectral data for the subobject, such as
                        "reflectance": reflectance,
                        "transmittance": transmittance,
                    },
                }
            }

        Each spectral data specified replaces the original data completely, so
        it is necessary to specify all the data for the object.

    Returns
    -------
    dict
        Returns a dictionary parsed from JSON with transformations applied.
    """

    # Load "scenario.json" as dictionary object
    scenario = json.loads((scenario_folder / SCENARIO_FILE_NAME).read_text())

    # Update dictionary elements with transformations
    surface = scenario["surface"]
    size = scenario["canopy"]["size"]
    center_2d = np.array([size[0], size[1], 0.0]) / 2
    return {
        **scenario,
        "surface": (
            _parse_rpv_surface(surface, spectral_data)
            if surface["type"] == "rpv"
            else _parse_lambertian_surface(surface, spectral_data)
        ),
        "canopy": {
            **scenario["canopy"],
            "instanced_canopy_elements": [
                {
                    **elem,
                    "instance_positions": [
                        apply_transformation(transf, center_2d)
                        for transf in elem["instance_positions"]
                    ],
                    "canopy_element": {
                        **elem["canopy_element"],
                        "mesh_tree_elements": [
                            _update_material(
                                {
                                    **tree,
                                    "mesh_filename": (
                                        scenario_folder / tree["mesh_filename"]
                                    ),
                                },
                                canopy_name=elem["canopy_element"]["id"],
                                spectral_data=spectral_data,
                            )
                            for tree in elem["canopy_element"]["mesh_tree_elements"]
                        ],
                    },
                }
                for elem in scenario["canopy"]["instanced_canopy_elements"]
            ],
            "padding": padding,
        },
    }



from __future__ import annotations

from contextlib import contextmanager

import xarray as xr

from ..typing import PathLike


def open_dataset(filename: PathLike, **kwargs) -> xr.Dataset:
    """
    Open a dataset from the online data store (requires Internet access).

    Parameters
    ----------
    filename : path-like
        Path to the requested dataset, relative to the data store root.

    **kwargs
        Forwarded to :func:`xarray.open_dataset`.

    Returns
    -------
    Dataset
        Open dataset, if found in the data store.

    See Also
    --------
    xarray.open_dataset
    """
    from ._store import data_store

    filename = data_store.fetch(filename)
    return xr.open_dataset(filename, **kwargs)


def load_dataset(*args, **kwargs) -> xr.Dataset:
    """
    Open, load into memory, and close a dataset from the online data store
    (requires Internet access).

    See Also
    --------
    open_dataset
    """
    with open_dataset(*args, **kwargs) as ds:
        return ds.load()


@contextmanager
def open_datasets(paths):
    try:
        datasets = [open_dataset(path) for path in paths]
        yield datasets
    finally:
        for ds in datasets:
            ds.close()



from __future__ import annotations

import enum
import errno
import glob
import json
import logging
import os
import re
import textwrap
import warnings
from collections.abc import Hashable, Mapping
from pathlib import Path
from typing import Any

import attrs
import numpy as np
import pandas as pd
import pint
import xarray as xr
from cachetools import LRUCache, cachedmethod
from pinttrs.util import ensure_units

import eradiate

from .. import config
from .._mode import SpectralMode
from ..attrs import documented, parse_docs
from ..data import data_store
from ..exceptions import InterpolationError, UnsupportedModeError
from ..typing import PathLike
from ..units import to_quantity
from ..units import unit_registry as ureg

logger = logging.getLogger(__name__)

# ------------------------------------------------------------------------------
#                           Error handling components
# ------------------------------------------------------------------------------


class ErrorHandlingAction(enum.Enum):
    IGNORE = "ignore"
    RAISE = "raise"
    WARN = "warn"


@attrs.define
class ErrorHandlingPolicy:
    missing: ErrorHandlingAction
    scalar: ErrorHandlingAction
    bounds: ErrorHandlingAction

    @classmethod
    def convert(cls, value):
        if isinstance(value, Mapping):
            kwargs = {k: ErrorHandlingAction(v) for k, v in value.items()}
            return cls(**kwargs)
        else:
            return value


@attrs.define
class ErrorHandlingConfiguration:
    x: ErrorHandlingPolicy = attrs.field(converter=ErrorHandlingPolicy.convert)
    p: ErrorHandlingPolicy = attrs.field(converter=ErrorHandlingPolicy.convert)
    t: ErrorHandlingPolicy = attrs.field(converter=ErrorHandlingPolicy.convert)

    @classmethod
    def convert(cls, value):
        if isinstance(value, dict):
            return cls(**value)
        else:
            return value


def handle_error(error: InterpolationError, action: ErrorHandlingAction):
    if action is ErrorHandlingAction.IGNORE:
        return

    if action is ErrorHandlingAction.WARN:
        warnings.warn(str(error), UserWarning)
        return

    if action is ErrorHandlingAction.RAISE:
        raise error

    raise NotImplementedError


ERROR_HANDLING_CONFIG_DEFAULT = ErrorHandlingConfiguration.convert(
    config.settings.ABSORPTION_DATABASE.ERROR_HANDLING
)

# ------------------------------------------------------------------------------
#                           Database type definitions
# ------------------------------------------------------------------------------


@parse_docs
@attrs.define(repr=False, eq=False)
class AbsorptionDatabase:
    """
    Common parent type for absorption coefficient databases.

    This class implements most of the data indexing logic common to all
    absorption coefficient databases.
    A database is composed of a set of NetCDF files compliant with the
    absorption coefficient database format specification and placed in the
    same directory. A database instance is initialized by specifying the path
    to the directory where the files are stored.

    If it exists, a ``metadata.json`` file is loaded into the :attr:`metadata`
    attribute.

    Databases are usually not initialized using the constructor, but rather
    using the class method constructors :meth:`from_directory`,
    :meth:`from_name` and :meth:`from_dict`.

    Notes
    -----
    A file index, stored as the :attr:`_index` private attribute, associates
    to each file the spectral region it covers. The index is preferably loaded
    from a CSV file that contains all this information; if it is not found, the
    table is built upon database initialization and saved to the database
    directory. The indexing step requires to access all files and may take a
    while. The file index table is used during queries to select efficiently the
    file where data will be read. For convenience, information about bounds
    contained in the index is assembled into a spectral mesh suitable for
    query using :func:`numpy.digitize` and stored in the :attr:`_chunks`
    dictionary.

    A spectral coverage table, stored as the :attr:`_spectral_coverage` private
    attribute, merges the spectral coordinates of all files into a consistent
    index. This table is used to provide spectral coverage information to
    higher-level components that drive the simulation. Table contents are
    preferably loaded from a CSV file; if it is not found, the table is build
    upon database initialization and saved to the database directory. This
    indexing step also requires to access all files and may take a while.

    Database access and memory usage can be controlled through two parameters:

    * File queries are stored in an LRU cache. The initial size is set to a low
      value (8) and should be appropriate for most situations. If more cache
      control is needed, the :meth:`cache_clear`,  :meth:`cache_close` and
      :meth:`cache_reset` methods can be used.
    * Datasets can be open with an eager or lazy approach. This behaviour is
      controlled using the ``lazy`` constructor parameter. In eager mode, the
      entire file used for a query is loaded into memory. This can bring
      significant access overhead when using large files. If desired, datasets
      can instead be open lazily, triggering disk access only for the specific
      data that are used.
    """

    _dir_path: Path = documented(
        attrs.field(converter=lambda x: Path(x).absolute().resolve()),
        doc="Path to database root directory.",
        type="pathlib.Path",
        init_type="path-like",
    )

    @_dir_path.validator
    def _dir_path_validator(self, attribute, value):
        if not value.is_dir():
            raise ValueError(
                f"while validating '{attribute.name}': path '{value}' is not a "
                "directory"
            )

    _index: pd.DataFrame = documented(
        attrs.field(repr=False),
        doc="File index, assumed sorted by ascending wavelengths.",
        type="DataFrame",
    )

    @_index.validator
    def _index_validator(self, attribute, value):
        if value.empty:
            raise ValueError(f"while validating '{attribute.name}': index is empty")

        wavelengths = value["wl_min [nm]"].values
        if not np.all(wavelengths[:-1] < wavelengths[1:]):
            raise ValueError(
                f"while validating '{attribute.name}': index must be sorted by "
                "ascending wavelength values"
            )

    _spectral_coverage: pd.DataFrame = documented(
        attrs.field(repr=False),
        doc="Dataframe that unrolls the spectral information contained in "
        "all data files in the database.",
        type="DataFrame",
    )

    _metadata: dict = documented(
        attrs.field(factory=dict, repr=False),
        doc="Dictionary that contains the database metadata.",
        type="dict",
    )

    _chunks: dict[str, np.ndarray] = documented(
        attrs.field(factory=dict, repr=False, init=False),
        doc="Dictionary mapping spectral lookup mode keys ('wl' or 'wn') to arrays "
        "containing the nodes of the spectral chunk mesh, which is used to perform "
        "spectral coordinate-based file lookup.",
        type="dict[str, ndarray]",
    )

    lazy: bool = documented(
        attrs.field(default=False, repr=False),
        doc="Access mode switch: if True, load data lazily; else, load data eagerly.",
        type="bool",
    )

    _cache: LRUCache = documented(
        attrs.field(factory=lambda: LRUCache(8), repr=False),
        doc="A mapping that implements an LRU caching policy.",
        type="cachetools.LRUCache",
    )

    def __attrs_post_init__(self):
        # Parse field names and units
        regex = re.compile(r"(?P<coord>.*)\_(?P<minmax>min|max) \[(?P<units>.*)\]")
        quantities = {}
        for colname in self._index.columns:
            if colname == "filename":
                continue

            m = regex.match(colname)
            units = m.group("units")
            magnitude = self._index[colname].values
            quantities[f"{m.group('coord')}_{m.group('minmax')}"] = ureg.Quantity(
                magnitude, units
            )

        # Populate spectral mesh (nodes) for both wavelength and wavenumber
        # lookup modes
        self._chunks["wl"] = np.concatenate(
            (quantities["wl_min"], [quantities["wl_max"][-1]])
        )
        self._chunks["wn"] = np.concatenate(
            (quantities["wn_max"], [quantities["wn_min"][-1]])
        )

    def __repr__(self) -> str:
        with pd.option_context("display.max_columns", 4):
            result = (
                f"<{type(self).__name__}> {self._dir_path}\n"
                f"Access mode: {'lazy' if self.lazy else 'eager'}\n"
                "Index:\n"
                f"{textwrap.indent(repr(self._index), '    ')}"
            )
        return result

    @staticmethod
    def _make_index(filenames) -> pd.DataFrame:
        headers = [
            "filename",
            "wn_min [cm^-1]",
            "wn_max [cm^-1]",
            "wl_min [nm]",
            "wl_max [nm]",
        ]
        rows = []

        for filename in filenames:
            filename = Path(filename)
            with xr.open_dataset(filename) as ds:
                w_u = ureg(ds.w.units)

                if w_u.check("[length]^-1"):  # wavenumber mode
                    wn_min = float(ds.w.min()) * w_u
                    wn_max = float(ds.w.max()) * w_u
                    wl_min = 1.0 / wn_max
                    wl_max = 1.0 / wn_min
                elif w_u.check("[length]"):  # wavelength mode
                    wl_min = float(ds.w.min()) * w_u
                    wl_max = float(ds.w.max()) * w_u
                    wn_min = 1.0 / wl_max
                    wn_max = 1.0 / wl_min
                else:
                    raise ValueError(f"Cannot interpret units '{w_u}'")

                rows.append(
                    [
                        filename.name,
                        wn_min.m_as("1/cm"),
                        wn_max.m_as("1/cm"),
                        wl_min.m_as("nm"),
                        wl_max.m_as("nm"),
                    ]
                )

        return pd.DataFrame(rows, columns=headers)

    @staticmethod
    def _make_spectral_coverage(filenames) -> pd.DataFrame:
        with xr.open_dataset(filenames[0]) as ds:
            dims = set(ds.dims)
            db_type = None
            if "w" in dims:
                db_type = "mono"
                if "g" in dims:
                    db_type = "ckd"

            if db_type is None:
                raise ValueError

            wavenumber_spectral_lookup_mode = ureg(ds.w.units).check("[length]^-1")

        index = []
        headers = ["wbound_lower [nm]", "wbound_upper [nm]"]
        rows = None

        for filename in filenames:
            filename = Path(filename)
            with xr.open_dataset(filename) as ds:
                w = to_quantity(ds.w)

                if wavenumber_spectral_lookup_mode:  # Convert to wavelength
                    w = 1.0 / w
                w = w.m_as("nm")

                if db_type == "mono":
                    wbounds_lower = np.full((len(w),), np.nan)
                    wbounds_upper = np.full((len(w),), np.nan)
                else:
                    wbounds_lower = to_quantity(ds.wbounds.sel(wbv="lower"))
                    wbounds_upper = to_quantity(ds.wbounds.sel(wbv="upper"))
                    if wavenumber_spectral_lookup_mode:  # Convert to wavelength
                        wbounds_lower = 1.0 / wbounds_lower
                        wbounds_upper = 1.0 / wbounds_upper
                    wbounds_lower = wbounds_lower.m_as("nm")
                    wbounds_upper = wbounds_upper.m_as("nm")

            index.extend([(filename.name, x) for x in w])

            if rows is None:
                rows = np.stack((wbounds_lower, wbounds_upper), axis=1)
            else:
                rows = np.concatenate(
                    (
                        rows,
                        np.stack((wbounds_lower, wbounds_upper), axis=1),
                    ),
                    axis=0,
                )

        index = pd.MultiIndex.from_tuples(index, names=["filename", "wavelength [nm]"])
        # Sort index by wavelength
        result = pd.DataFrame(rows, index=index, columns=headers).sort_index(level=1)
        return result

    @classmethod
    def from_directory(
        cls, dir_path: PathLike, lazy: bool = False, fix: bool = True
    ) -> AbsorptionDatabase:
        """
        Initialize a CKD database from a directory that contains one or several
        datasets.

        Parameters
        ----------
        dir_path : path-like
            Path where the CKD database is located.

        lazy : bool
            Access mode switch: if True, load data lazily; else, load data
            eagerly.

        fix : bool
            If ``True``, attempt generating missing index files upon
            initialization. Otherwise, raise if they are missing.

        Returns
        -------
        AbsorptionDatabase

        Raises
        ------
        FileNotFoundError
            If an index file is missing and ``fix`` is ``False``.
        """
        dir_path = Path(dir_path).resolve()

        try:
            with open(os.path.join(dir_path, "metadata.json")) as f:
                metadata = json.load(f)
        except FileNotFoundError:
            metadata = {}

        filenames = glob.glob(os.path.join(dir_path, "*.nc"))

        logger.debug("Loading index file")
        index_path = dir_path / "index.csv"
        if index_path.is_file():
            index = pd.read_csv(index_path)
        elif fix:
            logger.warning("Could not find index file, building it")
            index = cls._make_index(filenames)
            index.to_csv(index_path, index=False)
        else:
            logger.critical(f"Could not find index file '{index_path}'")
            raise FileNotFoundError(errno.ENOENT, "Missing index file", index_path)
        index = index.sort_values(by="wl_min [nm]").reset_index(drop=True)

        logger.debug("Loading spectral coverage table file")
        spectral_coverage_path = dir_path / "spectral.csv"
        if spectral_coverage_path.is_file():
            spectral_coverage = pd.read_csv(spectral_coverage_path, index_col=(0, 1))
        elif fix:
            logger.warning("Could not find spectral coverage table, building it")
            spectral_coverage = cls._make_spectral_coverage(filenames)
            spectral_coverage.to_csv(dir_path / "spectral.csv")
        else:
            logger.critical(
                f"Could not find spectral coverage table '{spectral_coverage_path}'"
            )
            raise FileNotFoundError(
                errno.ENOENT, "Missing spectral coverage table", spectral_coverage_path
            )

        return cls(dir_path, index, spectral_coverage, metadata=metadata, lazy=lazy)

    @staticmethod
    def from_name(name: str, **kwargs) -> AbsorptionDatabase:
        """
        Initialize a database from a name.

        Parameters
        ----------
        name : path-like
            Name of the requested CKD database.

        kwargs
            Additional keyword arguments are forwarded to the :meth:`.from_directory`
            constructor.

        Returns
        -------
        AbsorptionDatabase
        """

        try:
            db = KNOWN_DATABASES[name]
        except KeyError as e:
            raise ValueError(f"Unknown absorption coefficient database '{name}'") from e

        # Not great, but a.t.m it's the nicest we can do
        path = data_store.fetch(f"{db['path']}/metadata.json").parent
        cls = db["cls"]
        kwargs = {**db.get("kwargs", {}), **kwargs}

        # If an operational mode is selected, we check if the user is instantiating
        # a DB type that is relevant to that mode
        if eradiate.mode() is not None:
            cls_mode = MODE_TO_DB[eradiate.mode().spectral_mode]

            if cls is not cls_mode:
                warnings.warn(
                    f"Loading '{name}' database as a {cls.__name__}, but the current "
                    f"active mode requires to use a {cls_mode.__name__}. You might want "
                    "to consider loading a different database."
                )

        return cls.from_directory(path, **kwargs)

    @classmethod
    def from_dict(cls, value: dict) -> AbsorptionDatabase:
        """
        Construct from a dictionary. The dictionary has a required entry ``"construct"``
        that specifies the constructor that will be used to instantiate the
        database. Additional entries are keyword arguments passed to the selected
        constructor.

        Parameters
        ----------
        value : dict
            Converted value.

        Returns
        -------
        AbsorptionDatabase
        """

        raise NotImplementedError

    @staticmethod
    def default() -> AbsorptionDatabase:
        """
        Return a default database, depending on the active mode.

        Defaults are as follows:

        * Monochromatic: ``"komodo"``
        * CKD: ``"monotropa"``

        Returns
        -------
        AbsorptionDatabase
        """
        if eradiate.mode().is_mono:
            return AbsorptionDatabase.from_name("komodo")
        elif eradiate.mode().is_ckd:
            return AbsorptionDatabase.from_name("monotropa")
        else:
            raise UnsupportedModeError(supported=["mono", "ckd"])

    @staticmethod
    def convert(value: Any) -> AbsorptionDatabase:
        """
        Attempt conversion of a value to an absorption database.

        Parameters
        ----------
        value
            The value for which conversion is attempted.

        Returns
        -------
        MonoAbsorptionDatabase or CKDAbsorptionDatabase

        Notes
        -----
        Conversion rules are as follows:

        * If ``value`` is a string, try converting using the :meth:`.from_name`
          constructor. Do not raise if this fails.
        * If ``value`` is a string or a path, try converting using the
          :meth:`.from_directory` constructor. The returned type is consistent
          with the active mode.
        * If ``value`` is a dict, try converting using the :meth:`.from_dict`
          constructor. The returned type is consistent with the active mode.
        * Otherwise, do not convert.
        """
        if isinstance(value, str):
            try:
                return AbsorptionDatabase.from_name(value)
            except ValueError:
                pass

        if isinstance(value, (str, Path, dict)):
            if eradiate.mode().is_mono:
                cls = MonoAbsorptionDatabase
            elif eradiate.mode().is_ckd:
                cls = CKDAbsorptionDatabase
            else:
                raise UnsupportedModeError(supported=["mono", "ckd"])

            if isinstance(value, (str, Path)):
                return cls.from_directory(value)

            if isinstance(value, dict):
                return cls.from_dict(value)

        # Legacy behaviour
        if isinstance(value, (tuple, list)) and isinstance(value[0], str):
            warnings.warn(
                "Initializing an atmospheric molecular absorption database "
                "from a tuple is deprecated: specifying the spectral range is "
                "no longer necessary. Passing the database name is now enough. "
                "Support for this syntax will be removed in a future version.",
                DeprecationWarning,
            )
            return AbsorptionDatabase.from_name(value[0])

        return value

    @property
    def dir_path(self) -> Path:
        """
        Returns
        -------
        Path
            Database root path.
        """
        return self._dir_path

    @property
    def metadata(self) -> dict:
        """
        Returns
        -------
        dict
            Database metadata.
        """
        return self._metadata

    @property
    def spectral_coverage(self) -> pd.DataFrame:
        """
        Returns
        -------
        DataFrame
            Spectral coverage table.
        """
        return self._spectral_coverage

    @cachedmethod(lambda self: self._cache)
    def load_dataset(self, fname: str) -> xr.Dataset:
        """
        Convenience method to load a dataset. This method is decorated with
        :func:`functools.lru_cache` with ``maxsize=1``, which limits the number
        of reload events when repeatedly querying the same file.

        The behaviour of this method is also affected by the ``lazy`` parameter:
        if ``lazy`` is ``False``, files are loaded eagerly with
        :func:`xarray.load_dataset`; if ``lazy`` is ``True``, files are loaded
        lazily with :func:`xarray.open_dataset`.

        Parameters
        ----------
        fname : str
            Name of the file that is to be loaded.

        Returns
        -------
        Dataset
        """
        path = self._dir_path / fname

        if self.lazy:
            logger.debug("Opening '%s'" % path)
            return xr.open_dataset(path)
        else:
            logger.debug("Loading '%s'" % path)
            return xr.load_dataset(path)

    def cache_clear(self) -> None:
        """
        Clear the cache.
        """
        self._cache.clear()

    def cache_close(self) -> None:
        """
        Close all cached datasets.
        """
        for value in self._cache.values():
            value.close()

    def cache_reset(self, maxsize: int) -> None:
        """
        Reset the cache with the specified maximum size.
        """
        self._cache.clear()
        self._cache = LRUCache(maxsize=maxsize)

    def lookup_filenames(self, /, **kwargs) -> list[str]:
        """
        Look up a filename in the index table from the coordinate values passed
        as keyword arguments.

        Parameters
        ----------
        wl : quantity or array-like, optional
            Wavelength (scalar or array, quantity or unitless). If passed as a
            unitless value, it is interpreted using the units of the wavelength
            chunk bounds.

        wn : quantity or array-like, optional
            Wavenumber (scalar or array, quantity or unitless). If passed as a
            unitless value, it is interpreted using the units of the wavenumber
            chunk bounds.

        Returns
        -------
        filenames : list of str
            Names of the successfully looked up files, relative to the database
            root directory.

        Raises
        ------
        ValueError
            If the requested spectral coordinate is out of bounds.

        Notes
        -----
        Depending on the specified keyword argument (``wl`` or ``wn``), the
        lookup will be performed in wavelength or wavenumber mode. Both are
        equivalent.
        """
        if len(kwargs) != 1:
            raise ValueError(
                "only one of the 'wl' and 'wn' keyword arguments is allowed"
            )
        lookup_mode, values = next(iter(kwargs.items()))
        chunks = self._chunks[lookup_mode]

        # Make sure that 'values' has the right units
        values = ensure_units(np.atleast_1d(values), chunks.units)

        # Perform bound check
        out_bound = (values < chunks.min()) | (values > chunks.max())
        if np.any(out_bound):
            # TODO: handle this error better?
            raise ValueError("out-of-bound spectral coordinate value")

        indexes = np.digitize(values.m_as(chunks.units), bins=chunks.magnitude) - 1
        return list(self._index["filename"].iloc[indexes])

    def lookup_datasets(self, /, **kwargs) -> list[xr.Dataset]:
        """
        Perform a dataset lookup based on the requested spectral coordinate.
        See :meth:`lookup_filenames` for the accepted arguments.
        """
        filenames = self.lookup_filenames(**kwargs)
        return [self.load_dataset(filename) for filename in filenames]

    def eval_sigma_a_mono(
        self,
        w: pint.Quantity,
        thermoprops: xr.Dataset,
        error_handling_config: ErrorHandlingConfiguration | None = None,
    ) -> xr.DataArray:
        """
        Compute the absorption coefficient given spectral coordinates and a
        thermophysical profile (mono variant).

        Parameters
        ----------
        w : quantity
            The wavelength for which the absorption coefficient is evaluated.

        thermoprops : Dataset
            The thermophysical profile for which the absorption coefficient is
            evaluated.

        error_handling_config : .ErrorHandlingConfiguration, optional
            The error handling policy applied if corrdinates are missing, do not
            have the appropriate dimension or are out of the dataset's bounds.
            If unset, the default policy specified by the
            ``absorption_dataset.error_handling`` setting is applied.

        Returns
        -------
        DataArray
            A data array containing the evaluated absorption coefficient as a
            function of the spectral coordinate and altitude.

        See Also
        --------
        :mod:`eradiate.config`
        """
        raise NotImplementedError

    def eval_sigma_a_ckd(
        self,
        w: pint.Quantity,
        g: float,
        thermoprops: xr.Dataset,
        error_handling_config: ErrorHandlingConfiguration | None = None,
    ) -> xr.DataArray:
        """
        Compute the absorption coefficient given spectral coordinates and a
        thermophysical profile (CKD variant).

        Parameters
        ----------
        w : quantity
            The wavelength for which the absorption coefficient is evaluated.

        g : float
            The g-point for which the absorption coefficient is evaluated.

        thermoprops : Dataset
            The thermophysical profile for which the absorption coefficient is
            evaluated.

        error_handling_config : .ErrorHandlingConfiguration, optional
            The error handling policy applied if corrdinates are missing, do not
            have the appropriate dimension or are out of the dataset's bounds.
            If unset, the default policy specified by the
            ``absorption_dataset.error_handling`` setting is applied.

        Returns
        -------
        DataArray
            A data array containing the evaluated absorption coefficient as a
            function of the spectral coordinate and altitude.

        See Also
        --------
        :mod:`eradiate.config`
        """
        raise NotImplementedError

    @staticmethod
    def _interp_thermophysical(
        ds: xr.Dataset,
        da: xr.DataArray,
        thermoprops: xr.Dataset,
        error_handling_config: ErrorHandlingConfiguration,
    ) -> tuple[xr.DataArray, list[Hashable]]:
        # Interpolate on temperature
        bounds_error = error_handling_config.t.bounds is ErrorHandlingAction.RAISE
        fill_value = None if bounds_error else 0.0  # TODO: use 2-element tuple?
        result = da.interp(
            t=thermoprops.t,
            kwargs={"bounds_error": bounds_error, "fill_value": fill_value},
        )

        # Interpolate on pressure
        bounds_error = error_handling_config.p.bounds is ErrorHandlingAction.RAISE
        fill_value = None if bounds_error else 0.0  # TODO: use 2-element tuple?
        result = result.interp(
            p=thermoprops.p,
            kwargs={"bounds_error": bounds_error, "fill_value": fill_value},
        )

        # Interpolate on concentrations

        # -- List requested species concentrations
        x_ds = [coord for coord in ds.coords if coord.startswith("x_")]
        x_ds_scalar = [coord for coord in x_ds if ds[coord].size == 1]
        x_ds_array = set(x_ds) - set(x_ds_scalar)

        # -- Select on scalar coordinates
        result = result.isel(**{x: 0 for x in x_ds_scalar})

        # -- Interpolate on array coordinates
        bounds_error = error_handling_config.x.bounds is ErrorHandlingAction.RAISE
        fill_value = None if bounds_error else 0.0  # TODO: use 2-element tuple?
        result = result.interp(
            thermoprops[x_ds_array],
            kwargs={"bounds_error": bounds_error, "fill_value": fill_value},
        )

        return result, x_ds


@attrs.define(repr=False, eq=False)
class MonoAbsorptionDatabase(AbsorptionDatabase):
    """
    Absorption coefficient database (monochromatic variant).

    See Also
    --------
    AbsorptionDatabase
    """

    @classmethod
    def from_dict(cls, value: dict) -> MonoAbsorptionDatabase:
        # Inherit docstring
        value = value.copy()
        constructor = getattr(cls, value.pop("construct"))
        return constructor(**value)

    def eval_sigma_a_mono(
        self,
        w: pint.Quantity,
        thermoprops: xr.Dataset,
        error_handling_config: ErrorHandlingConfiguration | None = None,
    ) -> xr.DataArray:
        # Inherit docstring

        if error_handling_config is None:
            error_handling_config = ERROR_HANDLING_CONFIG_DEFAULT

        # Lookup dataset
        ds = self.lookup_datasets(wl=w)[0]

        # Interpolate on spectral dimension
        # TODO: Optimize
        w_u = ureg(ds.w.units)
        # Note: Support for wavenumber spectral lookup mode is suboptimal
        w_m = (1.0 / w).m_as(w_u) if w_u.check("[length]^-1") else w.m_as(w_u)
        result = ds.sigma_a.interp(w=w_m, method="linear")

        # Interpolate on thermophysical dimensions
        result, x_ds = self._interp_thermophysical(
            ds, result, thermoprops, error_handling_config
        )

        # Drop thermophysical coordinates, ensure spectral dimension
        result = result.drop_vars(["p", "t", *x_ds])
        if "w" not in result.dims:
            result = result.expand_dims("w")

        return result.transpose("w", "z")


@attrs.define(repr=False, eq=False)
class CKDAbsorptionDatabase(AbsorptionDatabase):
    """
    Absorption coefficient database (CKD variant).

    See Also
    --------
    AbsorptionDatabase
    """

    @classmethod
    def from_dict(cls, value: dict) -> CKDAbsorptionDatabase:
        # Inherit docstring
        value = value.copy()
        constructor = getattr(cls, value.pop("construct"))
        return constructor(**value)

    def eval_sigma_a_ckd(
        self,
        w: pint.Quantity,
        g: float,
        thermoprops: xr.Dataset,
        error_handling_config: ErrorHandlingConfiguration | None = None,
    ) -> xr.DataArray:
        # Inherit docstring

        # TODO: Implement new bounds error handling policy. This policy is as
        #  follows:
        #  * Interpolation is done for an altitude range such that the pressure
        #    is higher than the lower bound of the pressure variable in the
        #    CKD table. This is implemented at a higher level (not here).
        #  * The default bound error handling policy for the pressure and
        #    temperature variables is 'extrapolate'.
        #  * Above the cut-off altitude, the profile is filled with zeros.
        #  Cut-off detection is implemented with pressure-based masking.

        # TODO: Use the 'assume_sorted' parameter of DataArray.interp()

        if error_handling_config is None:
            error_handling_config = ERROR_HANDLING_CONFIG_DEFAULT

        # Lookup dataset
        ds = self.lookup_datasets(wl=w)[0]

        # Select bin
        # TODO: Optimize
        w_u = ureg(ds.w.units)
        w_m = w.m_as(w_u)
        result = ds.sigma_a.sel(w=w_m, method="nearest")

        # Interpolate along g
        result = result.interp(g=g).drop_vars("g")

        # Interpolate on thermophysical dimensions
        result, x_ds = self._interp_thermophysical(
            ds, result, thermoprops, error_handling_config
        )

        # Drop thermophysical coordinates, ensure spectral dimension
        result = result.drop_vars(["p", "t", *x_ds])
        if "w" not in result.dims:
            result = result.expand_dims("w")

        return result.transpose("w", "z")


# ------------------------------------------------------------------------------
#                               Static definitions
# ------------------------------------------------------------------------------

KNOWN_DATABASES = {
    "gecko": {
        "cls": MonoAbsorptionDatabase,
        "path": "spectra/absorption/mono/gecko",
        "kwargs": {"lazy": True},
    },
    "komodo": {
        "cls": MonoAbsorptionDatabase,
        "path": "spectra/absorption/mono/komodo",
        "kwargs": {"lazy": True},
    },
    "monotropa": {
        "cls": CKDAbsorptionDatabase,
        "path": "spectra/absorption/ckd/monotropa",
    },
    "mycena": {
        "cls": CKDAbsorptionDatabase,
        "path": "spectra/absorption/ckd/mycena",
    },
    "panellus": {
        "cls": CKDAbsorptionDatabase,
        "path": "spectra/absorption/ckd/panellus",
    },
}

MODE_TO_DB = {
    SpectralMode.MONO: MonoAbsorptionDatabase,
    SpectralMode.CKD: CKDAbsorptionDatabase,
}
